#!/usr/bin/python3 -u
# SPDX-License-Identifier: BSD-2

import unittest
from tpm2_pytss import *

try:
    from tpm2_pytss.TCTISPIHelper import TCTISPIHelper
except NotImplementedError:
    raise unittest.SkipTest("TCTISPIHelper not supported, skipping.")


class MySPITcti(TCTISPIHelper):

    TPM_DID_VID_HEAD = 0
    TPM_DID_VID_BODY = 1
    TPM_ACCESS_HEAD = 2
    TPM_ACCESS_BODY = 3
    TPM_STS_HEAD = 4
    TPM_STS_BODY = 5
    TPM_RID_HEAD = 6
    TPM_RID_BODY = 7

    TPM_DID_VID_0 = b"\x83\xd4\x0f\00\xd1\x15\x1b\x00"
    TPM_ACCESS_0 = b"\x80\xd4\x00\x00\xa1"
    TPM_STS_0 = b"\x83\xd4\x00\x18\x40\x00\x00\x00"
    TPM_RID_0 = b"\x80\xd4\x0f\x04\x00"

    def __init__(self, *args, with_exception=None, **kwargs):
        # Example of setting userdata, can be anything. You want
        # bound to this class, you can even add arguments, etc.
        self.with_exception = with_exception

        self.tpm_state = MySPITcti.TPM_DID_VID_HEAD

        # call the superclass
        super().__init__(*args, **kwargs)

    def on_sleep_ms(self, milliseconds: int) -> None:
        pass

    def on_start_timeout(self, milliseconds: int) -> None:
        pass

    def on_timeout_expired(self) -> bool:
        return True

    def on_spi_transfer(self, data_in: bytes) -> bytes:

        if self.with_exception:
            raise self.with_exception

        if self.waitstate:

            if self.tpm_state == MySPITcti.TPM_DID_VID_HEAD:
                b = bytearray(MySPITcti.TPM_DID_VID_0[:4])
                b[3] |= 0x01

            elif self.tpm_state == MySPITcti.TPM_DID_VID_BODY:
                b = MySPITcti.TPM_DID_VID_0[4:]

            elif self.tpm_state == MySPITcti.TPM_ACCESS_HEAD:
                b = bytearray(MySPITcti.TPM_ACCESS_0[:4])
                b[3] |= 0x01

            elif self.tpm_state == MySPITcti.TPM_ACCESS_BODY:
                b = MySPITcti.TPM_ACCESS_0[4:]

            elif self.tpm_state == MySPITcti.TPM_STS_HEAD:
                b = bytearray(MySPITcti.TPM_STS_0[:4])
                b[3] |= 0x01

            elif self.tpm_state == MySPITcti.TPM_STS_BODY:
                b = MySPITcti.TPM_STS_0[4:]

            elif self.tpm_state == MySPITcti.TPM_RID_HEAD:
                b = bytearray(MySPITcti.TPM_RID_0[:4])
                b[3] |= 0x01
            elif self.tpm_state == MySPITcti.TPM_RID_BODY:
                b = MySPITcti.TPM_RID_0[4:]
            else:
                raise RuntimeError("BAD STATE")

            self.tpm_state += 1

        else:
            if self.tpm_state == MySPITcti.TPM_DID_VID_HEAD:
                b = MySPITcti.TPM_DID_VID_0
            elif self.tpm_state == MySPITcti.TPM_ACCESS_HEAD:
                b = MySPITcti.TPM_ACCESS_0
            elif self.tpm_state == MySPITcti.TPM_STS_HEAD:
                b = MySPITcti.TPM_STS_0
            elif self.tpm_state == MySPITcti.TPM_RID_HEAD:
                b = MySPITcti.TPM_RID_0
            else:
                raise RuntimeError("BAD STATE")

            self.tpm_state += 2

        return bytes(b)

    def on_spi_acquire(self) -> None:
        pass

    def on_spi_release(self) -> None:
        pass


class MyBadSPITcti(TCTISPIHelper):
    def on_sleep_ms(self, milliseconds: int) -> None:
        pass

    def on_start_timeout(self, milliseconds: int) -> None:
        pass

    def on_timeout_expired(self) -> bool:
        return True

    def on_spi_transfer(self, data_in: bytes) -> bytes:
        pass


class MyBadSPITcti2(TCTISPIHelper):
    def on_sleep_ms(self, milliseconds: int) -> None:
        pass

    def on_start_timeout(self, milliseconds: int) -> None:
        pass

    def on_timeout_expired(self) -> bool:
        return True

    def on_spi_transfer(self, data_in: bytes) -> bytes:
        pass

    def on_spi_acquire(self) -> None:
        pass


class TestTCTI(unittest.TestCase):
    def test_spi_helper_good(self):
        MySPITcti()

    def yest_spi_helper_good_wait_state(self):
        MySPITcti(with_wait_state=True)

    def test_MyBadSPITcti(self):
        with self.assertRaises(NotImplementedError):
            MyBadSPITcti(with_wait_state=True)

    def test_MyBadSPITcti2(self):
        with self.assertRaises(NotImplementedError):
            MyBadSPITcti2(with_wait_state=True)

    def test_init_baseclase(self):
        with self.assertRaises(NotImplementedError):
            TCTISPIHelper()

    def test_init_baseclase_with_wait_state(self):
        with self.assertRaises(NotImplementedError):
            TCTISPIHelper(with_wait_state=True)

    def test_with_exception_accross_c(self):
        with self.assertRaises(RuntimeError, msg="foobar"):
            MySPITcti(with_exception=RuntimeError("foobar"))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3 -u
# SPDX-License-Identifier: BSD-2

import unittest
import textwrap
import json
from tpm2_pytss import *
from tpm2_pytss.internal.utils import _lib_version_atleast
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature
from .TSS2_BaseTest import TSS2_EsapiTest

if not _lib_version_atleast("tss2-policy", "4.0.0"):
    raise unittest.SkipTest("tss2-policy not installed or version is less then 4.0.0")


def lowercase_dict(src):
    if not isinstance(src, dict):
        return src
    dest = dict()
    for k, v in src.items():
        if isinstance(v, str):
            lv = v.lower()
            dest[k] = lv
        elif isinstance(v, dict):
            dest[k] = lowercase_dict(v)
        elif isinstance(v, list):
            l = list()
            for e in v:
                if isinstance(e, str):
                    le = e.lower()
                elif isinstance(e, dict):
                    le = lowercase_dict(e)
                else:
                    le = e
                l.append(le)
            dest[k] = l
        else:
            dest[k] = v
    return dest


class TestPolicy(TSS2_EsapiTest):
    def setUp(self):
        super().setUp()
        self._has_secp192r1 = True
        try:
            ec.generate_private_key(ec.SECP192R1())
        except Exception:
            self._has_secp192r1 = False

    def test_password_policy(self):
        pol = {
            "description": "this is a policy",
            "policy": [{"type": "POLICYPASSWORD"}],
        }
        calcpol = {
            "description": "this is a policy",
            "policyDigests": [
                {
                    "digest": "af6038c78c5c962d37127e319124e3a8dc582e9b",
                    "hashAlg": "SHA1",
                }
            ],
            "policy": [
                {
                    "type": "POLICYPASSWORD",
                    "policyDigests": [
                        {
                            "digest": "af6038c78c5c962d37127e319124e3a8dc582e9b",
                            "hashAlg": "SHA1",
                        }
                    ],
                }
            ],
        }

        polstr = json.dumps(pol).encode()
        with policy(polstr, TPM2_ALG.SHA1) as p:
            desc = p.description
            polp = p.policy
            halg = p.hash_alg
            p.calculate()
            dig = p.get_calculated_digest()
            cjb = p.get_calculated_json()
        self.assertEqual(desc, b"this is a policy")
        self.assertEqual(polp, polstr)
        self.assertEqual(halg, TPM2_ALG.SHA1)
        self.assertEqual(
            dig.buffer, b"\xaf`8\xc7\x8c\\\x96-7\x12~1\x91$\xe3\xa8\xdcX.\x9b"
        )
        cj = json.loads(cjb)
        self.assertEqual(lowercase_dict(cj), lowercase_dict(calcpol))

    def test_password_policy_execute(self):
        pol = {
            "description": "this is a policy",
            "policy": [{"type": "POLICYPASSWORD"}],
        }
        polstr = json.dumps(pol).encode()
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA1,
        )
        with policy(polstr, TPM2_ALG.SHA1) as p:
            p.execute(self.ectx, session)
        dig2b = self.ectx.policy_get_digest(session)
        self.assertEqual(
            dig2b.buffer, b"\xaf`8\xc7\x8c\\\x96-7\x12~1\x91$\xe3\xa8\xdcX.\x9b"
        )

    def test_callbacks(self):
        pol = {
            "description": "this is a policy",
            "policy": [{"type": "POLICYPASSWORD"}],
        }
        polstr = json.dumps(pol)

        def test():
            pass

        p = policy(polstr, TPM2_ALG.SHA256)
        p.set_callback(policy_cb_types.CALC_PCR, test)
        cb = p._get_callback(policy_cb_types.CALC_PCR)
        self.assertEqual(cb, test)

        p.set_callback(policy_cb_types.CALC_PCR, None)
        cb = p._get_callback(policy_cb_types.CALC_PCR)
        self.assertEqual(cb, None)

        with self.assertRaises(ValueError) as e:
            p.set_callback(1234, test)
        self.assertEqual(str(e.exception), "unsupported callback type 1234")

    def test_calc_pcr_callback(self):
        pol = {
            "description": "this is a pcr policy",
            "policy": [{"type": "pcr", "currentPCRs": [0, 8]}],
        }
        calc_pol = {
            "description": "this is a pcr policy",
            "policyDigests": [
                {
                    "hashAlg": "SHA256",
                    "digest": "e5a442791c55f8c4a3e391385e170a24c75add21bc2c140fc4f4a2628810f13f",
                }
            ],
            "policy": [
                {
                    "type": "POLICYPCR",
                    "policyDigests": [
                        {
                            "hashAlg": "SHA256",
                            "digest": "e5a442791c55f8c4a3e391385e170a24c75add21bc2c140fc4f4a2628810f13f",
                        }
                    ],
                    "pcrs": [
                        {"pcr": 0, "hashAlg": "SHA256", "digest": "00" * 32},
                        {"pcr": 8, "hashAlg": "SHA256", "digest": "08" * 32},
                    ],
                }
            ],
        }
        polstr = json.dumps(pol).encode()
        cb_called = False
        cb_selection = TSS2_POLICY_PCR_SELECTION()

        def pcr_cb(selection):
            nonlocal cb_called
            nonlocal cb_selection
            cb_called = True
            cb_selection = selection
            sel = TPMS_PCR_SELECTION(
                hash=TPM2_ALG.SHA256,
                sizeofSelect=selection.selections.pcr_select.sizeofSelect,
                pcrSelect=selection.selections.pcr_select.pcrSelect,
            )
            out_sel = TPML_PCR_SELECTION((sel,))
            digests = list()
            selb = bytes(sel.pcrSelect[0 : sel.sizeofSelect])
            seli = int.from_bytes(reversed(selb), "big")
            for i in range(0, sel.sizeofSelect * 8):
                if (1 << i) & seli:
                    dig = TPM2B_DIGEST(bytes([i]) * 32)
                    digests.append(dig)
            out_dig = TPML_DIGEST(digests)
            return (out_sel, out_dig)

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_PCR, pcr_cb)
            p.calculate()
            cjb = p.get_calculated_json()

        cj = json.loads(cjb)
        self.assertTrue(cb_called)
        self.assertIsInstance(cb_selection, TSS2_POLICY_PCR_SELECTION)
        self.assertEqual(cb_selection.type, TSS2_POLICY_PCR_SELECTOR.PCR_SELECT)
        self.assertEqual(cb_selection.selections.pcr_select.sizeofSelect, 3)
        self.assertEqual(
            bytes(cb_selection.selections.pcr_select.pcrSelect), b"\x01\x01\x00\x00"
        )
        self.assertEqual(lowercase_dict(cj), lowercase_dict(calc_pol))

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_PCR, bad_cb)
            p.calculate()
        self.assertEqual(str(e.exception), "callback exception")

    def test_calc_name_callback(self):
        pol = {
            "description": "this is a name policy",
            "policy": [{"type": "namehash", "namePaths": ["path1", "path2", "path3"]}],
        }
        calc_pol = {
            "description": "this is a name policy",
            "policyDigests": [
                {
                    "hashAlg": "SHA256",
                    "digest": "7334f60c505d0007343c33e697cadce69ca56cebe7870ad68c65534b1eee5fa6",
                }
            ],
            "policy": [
                {
                    "type": "POLICYNAMEHASH",
                    "policyDigests": [
                        {
                            "hashAlg": "SHA256",
                            "digest": "7334f60c505d0007343c33e697cadce69ca56cebe7870ad68c65534b1eee5fa6",
                        }
                    ],
                    "nameHash": "15c6f51a0c7d0b68942d62b4eaf59b3240e819a566361b0b052f46e4422051d8",
                }
            ],
        }
        polstr = json.dumps(pol).encode()

        cb_called = 0
        cb_names = list()

        def name_cb(name):
            nonlocal cb_called
            nonlocal cb_names
            cb_called += 1
            cb_names.append(name)
            return TPM2B_NAME(b"\x00\x0b" + bytes([cb_called]) * 32)

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_NAME, name_cb)
            p.calculate()
            cjb = p.get_calculated_json()

        cj = json.loads(cjb)
        self.assertEqual(cb_called, 3)
        self.assertEqual(cb_names, [b"path1", b"path2", b"path3"])
        self.assertEqual(lowercase_dict(cj), lowercase_dict(calc_pol))

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_NAME, bad_cb)
            p.calculate()
        self.assertEqual(str(e.exception), "callback exception")

    def test_calc_public_callback(self):
        pol = {
            "description": "this is a public policy",
            "policy": [{"type": "duplicationselect", "newParentPath": "parent_path"}],
        }
        calc_pol = {
            "description": "this is a public policy",
            "policyDigests": [
                {
                    "hashAlg": "SHA256",
                    "digest": "fd6ba57a0029f93e76628ecffe78911df5c93cc24e5d53e06472e0c3dd116e76",
                }
            ],
            "policy": [
                {
                    "type": "POLICYDUPLICATIONSELECT",
                    "policyDigests": [
                        {
                            "hashAlg": "SHA256",
                            "digest": "fd6ba57a0029f93e76628ecffe78911df5c93cc24e5d53e06472e0c3dd116e76",
                        }
                    ],
                    "objectName": "",
                    "newParentName": "000b6e59ef657bca3b9624088244e6f67ab28d7843a8f53df58e37c67b0df7152f44",
                    "includeObject": "NO",
                },
            ],
        }
        polstr = json.dumps(pol).encode()

        cb_path = None

        def public_cb(path):
            nonlocal cb_path
            cb_path = path
            return TPMT_PUBLIC.parse("rsa2048")

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_PUBLIC, public_cb)
            p.calculate()
            cjb = p.get_calculated_json()

        cj = json.loads(cjb)
        self.assertEqual(cb_path, b"parent_path")
        self.assertEqual(lowercase_dict(cj), lowercase_dict(calc_pol))

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_PUBLIC, bad_cb)
            p.calculate()
        self.assertEqual(str(e.exception), "callback exception")

    def test_calc_nvpublic_callback(self):
        pol = {
            "description": "this is a nvpublic policy",
            "policy": [{"type": "authorizenv", "nvPath": "nv_path"}],
        }
        calc_pol = {
            "description": "this is a nvpublic policy",
            "policyDigests": [
                {
                    "hashAlg": "SHA256",
                    "digest": "31e4f72a6ca046ca0dc4a8765b400ac01fe76502b9b820e7927a50b253624494",
                }
            ],
            "policy": [
                {
                    "type": "POLICYAUTHORIZENV",
                    "policyDigests": [
                        {
                            "hashAlg": "SHA256",
                            "digest": "31e4f72a6ca046ca0dc4a8765b400ac01fe76502b9b820e7927a50b253624494",
                        }
                    ],
                    "nvPublic": {
                        "nvIndex": 0x1000000,
                        "nameAlg": "SHA256",
                        "attributes": {
                            "PPWRITE": 0,
                            "OWNERWRITE": 0,
                            "AUTHWRITE": 0,
                            "POLICYWRITE": 0,
                            "POLICY_DELETE": 0,
                            "WRITELOCKED": 0,
                            "WRITEALL": 0,
                            "WRITEDEFINE": 0,
                            "WRITE_STCLEAR": 0,
                            "GLOBALLOCK": 0,
                            "PPREAD": 0,
                            "OWNERREAD": 0,
                            "AUTHREAD": 0,
                            "POLICYREAD": 0,
                            "NO_DA": 0,
                            "ORDERLY": 0,
                            "CLEAR_STCLEAR": 0,
                            "READLOCKED": 0,
                            "WRITTEN": 1,
                            "PLATFORMCREATE": 0,
                            "READ_STCLEAR": 0,
                            "TPM2_NT": "ORDINARY",
                        },
                        "authPolicy": "",
                        "dataSize": 0,
                    },
                },
            ],
        }
        polstr = json.dumps(pol).encode()

        cb_nvpath = None
        cb_nvindex = None

        def nvpublic_cb(path, index):
            nonlocal cb_nvpath, cb_nvindex
            cb_nvpath = path
            cb_nvindex = index
            nvp = TPMS_NV_PUBLIC(nvIndex=0x1000000, nameAlg=TPM2_ALG.SHA256)
            return nvp

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_NVPUBLIC, nvpublic_cb)
            p.calculate()
            cjb = p.get_calculated_json()

        cj = json.loads(cjb)
        self.assertEqual(cb_nvpath, b"nv_path")
        self.assertEqual(cb_nvindex, 0)
        self.assertIsInstance(cb_nvindex, TPM2_HANDLE)
        self.assertEqual(lowercase_dict(cj), lowercase_dict(calc_pol))

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_NVPUBLIC, bad_cb)
            p.calculate()
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_auth_callback(self):
        pol = {
            "description": "this is an auth policy",
            "policy": [{"type": "nv", "nvIndex": "0x1000000", "operandB": "00"}],
        }
        polstr = json.dumps(pol).encode()
        nvp = TPMS_NV_PUBLIC(
            nvIndex=0x1000000,
            nameAlg=TPM2_ALG.SHA256,
            attributes=TPMA_NV.AUTHWRITE | TPMA_NV.AUTHREAD,
            dataSize=1,
        )
        nvh = self.ectx.nv_define_space(None, TPM2B_NV_PUBLIC(nvp))
        self.ectx.nv_write(nvh, b"\x00")

        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA1,
        )

        def nvpublic_cb(path, index):
            nonlocal nvp
            if index == 0x1000000:
                return nvp
            return None

        cb_name = TPM2B_NAME()

        def auth_cb(name):
            nonlocal cb_name
            nonlocal nvh
            cb_name = name
            return (nvh, nvh, ESYS_TR.PASSWORD)

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_NVPUBLIC, nvpublic_cb)
            p.set_callback(policy_cb_types.EXEC_AUTH, auth_cb)
            p.execute(self.ectx, session)

        dig2b = self.ectx.policy_get_digest(session)
        self.assertEqual(
            cb_name,
            b"\x00\x0b\x11\xf2\x07J\x1e\xcde\xf8T\x1cU\x15T\x80\xbb<\xdb\x83\xbf\x10\xffy\x8b\x9bB\\n\xa8E\xbe\xeaP",
        )
        self.assertEqual(
            dig2b, b"Ad\x9eq\xa0!\xc0\xef\xe8v\x053\x97x\xbd;\xfa\x9c\x14|"
        )

        def bad_cb(name):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_NVPUBLIC, nvpublic_cb)
            p.set_callback(policy_cb_types.EXEC_AUTH, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_polsel_callback(self):
        pol = {
            "description": "this is a polsel policy",
            "policy": [
                {
                    "type": "or",
                    "branches": [
                        {
                            "name": "branch1",
                            "description": "branch1 description",
                            "policy": [{"type": "password"}],
                        },
                        {
                            "name": "branch2",
                            "description": "branch2 description",
                            "policy": [{"type": "locality", "locality": ["zero"]}],
                        },
                    ],
                },
            ],
        }
        polstr = json.dumps(pol).encode()
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA256,
        )
        cb_auth_object = ""
        cb_branches = None

        def polsel_cb(auth_object, branches):
            nonlocal cb_auth_object
            nonlocal cb_branches
            cb_auth_object = auth_object
            cb_branches = branches
            return len(branches) - 1

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLSEL, polsel_cb)
            p.execute(self.ectx, session)

        dig2b = self.ectx.policy_get_digest(session)
        self.assertEqual(cb_auth_object, None)
        self.assertEqual(cb_branches, [b"branch1", b"branch2"])
        self.assertEqual(
            dig2b,
            b"8\x17\xfa\x84\x98\xf9\xf6\xa0s\xaa\xc5\x91r\x0b\xc4\xea\xdf3\xd6\xdb#\xd5\n\x05\x12\xd7\x8a\x84\xb5\xa3\xb2\xa1",
        )

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLSEL, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_sign_callback(self):
        if not self._has_secp192r1:
            self.skipTest("cryptography doesn't have secp192r1")
        private_key = textwrap.dedent(
            """
            -----BEGIN PRIVATE KEY-----
            MG8CAQAwEwYHKoZIzj0CAQYIKoZIzj0DAQEEVTBTAgEBBBgBJYQqvoPfXctJixFL
            lAzRLQaAFBHOoQyhNAMyAATAjqP6LEx2q1p5aUSAfSwIpr0NijnvyLfWtluYrqCJ
            sI7HNirP/FKiz8pIY3FAD18=
            -----END PRIVATE KEY-----
            """
        ).encode("ascii")
        public_key = textwrap.dedent(
            """
            -----BEGIN PUBLIC KEY-----
            MEkwEwYHKoZIzj0CAQYIKoZIzj0DAQEDMgAEwI6j+ixMdqtaeWlEgH0sCKa9DYo5
            78i31rZbmK6gibCOxzYqz/xSos/KSGNxQA9f
            -----END PUBLIC KEY-----
            """
        ).encode("ascii")
        pol = {
            "description": "this is a sign policy",
            "policy": [
                {
                    "type": "signed",
                    "publicKeyHint": "test key",
                    "keyPEM": public_key.decode(),
                },
            ],
        }
        polstr = json.dumps(pol).encode()
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA256,
        )
        pkey = load_pem_private_key(private_key, password=None)
        cb_pem = None
        cb_key_hint = None
        cb_hash_alg = None
        cb_buf = None

        def sign_cb(pem, key_hint, hash_alg, buf):
            nonlocal pkey, cb_pem, cb_key_hint, cb_hash_alg, cb_buf
            cb_pem = pem
            cb_key_hint = key_hint
            cb_hash_alg = hash_alg
            cb_buf = buf
            sig = pkey.sign(buf, ec.ECDSA(hashes.SHA256()))
            return sig

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_SIGN, sign_cb)
            p.execute(self.ectx, session)

        dig2b = self.ectx.policy_get_digest(session)
        self.assertEqual(cb_pem, public_key.lstrip(b"\n"))
        self.assertEqual(cb_key_hint, b"test key")
        self.assertEqual(cb_hash_alg, TPM2_ALG.SHA256)
        self.assertEqual(
            dig2b,
            b"\xc6L!\x81v\x893\xee\x14)^eYi/\xa3\x88\xa6}\xbf\xf1\x86\x99\x90\x9e\x10fg99\xecL",
        )

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_SIGN, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_polauth_callback(self):
        if not self._has_secp192r1:
            self.skipTest("cryptography doesn't have secp192r1")
        private_key = textwrap.dedent(
            """
            -----BEGIN PRIVATE KEY-----
            MG8CAQAwEwYHKoZIzj0CAQYIKoZIzj0DAQEEVTBTAgEBBBgBJYQqvoPfXctJixFL
            lAzRLQaAFBHOoQyhNAMyAATAjqP6LEx2q1p5aUSAfSwIpr0NijnvyLfWtluYrqCJ
            sI7HNirP/FKiz8pIY3FAD18=
            -----END PRIVATE KEY-----
            """
        ).encode("ascii")
        public_key = textwrap.dedent(
            """
            -----BEGIN PUBLIC KEY-----
            MEkwEwYHKoZIzj0CAQYIKoZIzj0DAQEDMgAEwI6j+ixMdqtaeWlEgH0sCKa9DYo5
            78i31rZbmK6gibCOxzYqz/xSos/KSGNxQA9f
            -----END PUBLIC KEY-----
            """
        ).encode("ascii")
        pol = {
            "description": "this is a polauth policy",
            "policy": [
                {
                    "type": "authorize",
                    "keyPEM": public_key.decode(),
                    "approvedPolicy": "01" * 32,
                    "policyRef": "02" * 32,
                },
            ],
        }
        polstr = json.dumps(pol).encode()
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.TRIAL,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA256,
        )
        pkey = load_pem_private_key(private_key, password=None)
        cb_hash_alg = None
        cb_digest = None
        cb_policyref = None

        def polauth_cb(key_public, hash_alg, digest, policy_ref):
            nonlocal pkey, cb_hash_alg, cb_digest, cb_policyref
            cb_hash_alg = hash_alg
            cb_digest = digest
            cb_policyref = policy_ref
            buf = bytes(digest) + bytes(policy_ref)
            sig = pkey.sign(buf, ec.ECDSA(hashes.SHA256()))
            r, s = decode_dss_signature(sig)
            rlen = int(r.bit_length() / 8) + (r.bit_length() % 8 > 0)
            slen = int(s.bit_length() / 8) + (s.bit_length() % 8 > 0)
            rb = r.to_bytes(rlen, "big")
            sb = s.to_bytes(slen, "big")
            tpm_sig = TPMT_SIGNATURE(
                sigAlg=TPM2_ALG.ECDSA,
                signature=TPMU_SIGNATURE(
                    ecdsa=TPMS_SIGNATURE_ECC(
                        hash=TPM2_ALG.SHA256, signatureR=rb, signatureS=sb
                    )
                ),
            )
            return tpm_sig

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLAUTH, polauth_cb)
            p.execute(self.ectx, session)

        self.assertEqual(cb_hash_alg, TPM2_ALG.SHA256)
        self.assertEqual(cb_digest, b"\x01" * 32)
        self.assertEqual(cb_policyref, b"\x02" * 32)

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLAUTH, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_polauthnv_callback(self):
        pol = {
            "description": "this is a polauthnv policy",
            "policy": [
                {
                    "type": "authorizenv",
                    "nvPublic": {
                        "nvIndex": 0x1000000,
                        "nameAlg": "SHA256",
                        "attributes": {
                            "PPWRITE": 0,
                            "OWNERWRITE": 0,
                            "AUTHWRITE": 1,
                            "POLICYWRITE": 0,
                            "POLICY_DELETE": 0,
                            "WRITELOCKED": 0,
                            "WRITEALL": 0,
                            "WRITEDEFINE": 0,
                            "WRITE_STCLEAR": 0,
                            "GLOBALLOCK": 0,
                            "PPREAD": 0,
                            "OWNERREAD": 0,
                            "AUTHREAD": 1,
                            "POLICYREAD": 0,
                            "NO_DA": 0,
                            "ORDERLY": 0,
                            "CLEAR_STCLEAR": 0,
                            "READLOCKED": 0,
                            "WRITTEN": 1,
                            "PLATFORMCREATE": 0,
                            "READ_STCLEAR": 0,
                            "TPM2_NT": "ORDINARY",
                        },
                        "authPolicy": "",
                        "dataSize": 34,
                    },
                },
            ],
        }
        polstr = json.dumps(pol)
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA256,
        )
        nvp = TPMS_NV_PUBLIC(
            nvIndex=0x1000000,
            nameAlg=TPM2_ALG.SHA256,
            attributes=TPMA_NV.AUTHWRITE | TPMA_NV.AUTHREAD,
            dataSize=34,
        )
        nvh = self.ectx.nv_define_space(None, TPM2B_NV_PUBLIC(nvp))
        self.ectx.nv_write(nvh, b"\x00\x0b" + b"\x00" * 32)

        cb_nv_public = None
        cb_hash_alg = None

        def polauthnv_cb(nv_public, hash_alg):
            nonlocal cb_nv_public, cb_hash_alg
            cb_nv_public = nv_public
            cb_hash_alg = hash_alg

        cb_name = None

        def auth_cb(name):
            nonlocal cb_name, nvh
            cb_name = name
            return (nvh, nvh, ESYS_TR.PASSWORD)

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLAUTHNV, polauthnv_cb)
            p.set_callback(policy_cb_types.EXEC_AUTH, auth_cb)
            p.execute(self.ectx, session)

        self.assertEqual(cb_nv_public.nvIndex, nvp.nvIndex)
        self.assertEqual(cb_nv_public.nameAlg, nvp.nameAlg)
        self.assertEqual(cb_nv_public.attributes, nvp.attributes | TPMA_NV.WRITTEN)
        self.assertEqual(cb_nv_public.dataSize, nvp.dataSize)
        self.assertEqual(cb_hash_alg, TPM2_ALG.SHA256)
        self.assertEqual(
            cb_name,
            b"\x00\x0bx\x7f\x8a\xa5&\xdbK\xf2L\x97\x8by\x92\x1f\xf4*\xae\xe6E\xa1\x15\xfb|\x05]\xed\xd4\x9f\xc3\xb5\xd1\xf6",
        )

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_AUTH, auth_cb)
            p.set_callback(policy_cb_types.EXEC_POLAUTHNV, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_poldup_callback(self):
        pol = {
            "description": "this is a poldup policy",
            "policy": [
                {"type": "duplicationselect", "newParentPath": "new parent path"},
            ],
        }
        polstr = json.dumps(pol)
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA256,
        )

        cb_path = None

        def poldup_cb():
            return TPM2B_NAME(b"\x12" * 32)

        def public_cb(path):
            nonlocal cb_path
            cb_path = path
            return TPMT_PUBLIC.parse("rsa2048")

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_PUBLIC, public_cb)
            p.set_callback(policy_cb_types.EXEC_POLDUP, poldup_cb)
            p.execute(self.ectx, session)
        dig2b = self.ectx.policy_get_digest(session)

        self.assertEqual(cb_path, b"new parent path")
        self.assertEqual(
            dig2b,
            b"\xfdk\xa5z\x00)\xf9>vb\x8e\xcf\xfex\x91\x1d\xf5\xc9<\xc2N]S\xe0dr\xe0\xc3\xdd\x11nv",
        )

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.CALC_PUBLIC, public_cb)
            p.set_callback(policy_cb_types.EXEC_POLDUP, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")

    def test_exec_polaction_callback(self):
        pol = {
            "description": "this is a polaction policy",
            "policy": [{"type": "action", "action": "this is an action"}],
        }
        polstr = json.dumps(pol)
        session = self.ectx.start_auth_session(
            ESYS_TR.NONE,
            ESYS_TR.NONE,
            TPM2_SE.POLICY,
            TPMT_SYM_DEF(algorithm=TPM2_ALG.NULL),
            TPM2_ALG.SHA256,
        )

        cb_action = None

        def polaction_cb(action):
            nonlocal cb_action
            cb_action = action

        with policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLACTION, polaction_cb)
            p.execute(self.ectx, session)

        self.assertEqual(cb_action, b"this is an action")

        def bad_cb(*args):
            raise Exception("callback exception")

        with self.assertRaises(Exception) as e, policy(polstr, TPM2_ALG.SHA256) as p:
            p.set_callback(policy_cb_types.EXEC_POLACTION, bad_cb)
            p.execute(self.ectx, session)
        self.assertEqual(str(e.exception), "callback exception")



#!/usr/bin/python3 -u
# SPDX-License-Identifier: BSD-2

import unittest

from tpm2_pytss import *
from .TSS2_BaseTest import TSS2_EsapiTest


class MyTCTI(PyTCTI):
    def __init__(self, subtcti, magic=None):
        self._tcti = subtcti
        self._is_finalized = False
        self._error = None

        if magic is not None:
            super().__init__(magic=magic)
        else:
            super().__init__()

    @property
    def is_finalized(self):
        return self._is_finalized

    def do_transmit(self, command):
        self._tcti.transmit(command)

    def do_receive(self, timeout):
        return self._tcti.receive()

    def do_cancel(self):
        self._tcti.cancel()

    def do_get_poll_handles(self):
        return self._tcti.get_poll_handles()

    def do_set_locality(self, locality):
        self._tcti.set_locality(locality)

    def do_make_sticky(self, handle, is_sticky):
        if self._tcti is not None:
            self._tcti.make_sticky(handle, is_sticky)

        if self._error is not None:
            raise self._error

    def do_finalize(self):
        self._is_finalized = True
        if self._error is not None:
            raise self._error


class TestTCTI(TSS2_EsapiTest):
    def test_init(self):
        self.assertEqual(self.tcti.version, 2)
        self.assertGreater(int.from_bytes(self.tcti.magic, "big"), 0)

        v1ctx = ffi.cast("TSS2_TCTI_CONTEXT_COMMON_V1 *", self.tcti._ctx)
        v1ctx.version = 1

        tcti = TCTI(self.tcti._ctx)
        self.assertEqual(tcti.version, 1)
        self.assertEqual(tcti._v2, None)

    def test_transmit_receive(self):
        startup = b"\x80\x01\x00\x00\x00\x0C\x00\x00\x01\x44\x00\x00"
        self.tcti.transmit(startup)

        resp = self.tcti.receive()
        self.assertEqual(resp, b"\x80\x01\x00\x00\x00\n\x00\x00\x01\x00")

    def test_finalize(self):
        tcti = TCTI(self.tcti._ctx)
        tcti.finalize()

    def test_cancel(self):
        if getattr(self.tcti, "name", "") == "swtpm":
            self.skipTest("cancel not supported by swtpm")

        startup = b"\x80\x01\x00\x00\x00\x0C\x00\x00\x01\x44\x00\x00"
        self.tcti.transmit(startup)
        self.tcti.cancel()

    def test_get_poll_handles(self):
        tcti_name = getattr(self.tcti, "name", "")
        try:
            self.tcti.get_poll_handles()
        except TSS2_Exception as e:
            if e.rc != lib.TSS2_TCTI_RC_NOT_IMPLEMENTED:
                raise e
            else:
                self.skipTest(f"get_poll_handles not supported by {tcti_name}")

    def test_set_locality(self):
        self.tcti.set_locality(TPMA_LOCALITY.TWO)

    def test_make_sticky(self):
        tcti_name = getattr(self.tcti, "name", "")
        if tcti_name in ("swtpm", "mssim"):
            self.skipTest(f"make_sticky not supported by {tcti_name}")

        self.tcti.make_sticky(0, 0)

        tcti._v2 = None
        with self.assertRaises(RuntimeError) as e:
            self.tcti.make_sticky(0, 0)
        self.assertEqual(str(e.exception), "unsupported by TCTI API version")

    def test_tctildr(self):
        self.assertIsInstance(self.tcti.name, str)
        self.assertIsInstance(self.tcti.conf, str)

        with self.assertRaises(TypeError):
            TCTILdr(name=None, conf=1234)

        with self.assertRaises(TypeError):
            TCTILdr(name=1234, conf=None)

    def test_custom_pytcti_esapi(self):

        t = MyTCTI(self.tcti)
        e = ESAPI(t)
        e.get_random(4)

        e.startup(TPM2_SU.CLEAR)

    def test_custom_pytcti_C_wrapper_transmit_receive(self):

        t = MyTCTI(self.tcti)

        # Go through the C API directly and call transmit and recv
        t.transmit(b"\x80\x01\x00\x00\x00\x0C\x00\x00\x01\x44\x00\x00")
        resp = t.receive(-1)
        self.assertEqual(resp, b"\x80\x01\x00\x00\x00\n\x00\x00\x01\x00")

    def test_custom_pytcti_cancel(self):
        if getattr(self.tcti, "name", "") == "swtpm":
            self.skipTest("cancel not supported by swtpm")

        t = MyTCTI(self.tcti)

        t.transmit(b"\x80\x01\x00\x00\x00\x0C\x00\x00\x01\x44\x00\x00")
        t.cancel()

    def test_custom_pytcti_finalize(self):
        t = MyTCTI(self.tcti)
        t.finalize()
        self.assertTrue(t.is_finalized)

    def test_custom_pytcti_get_poll_handles(self):
        tcti_name = getattr(self.tcti, "name", "")
        t = MyTCTI(self.tcti)
        try:
            handles = t.get_poll_handles()
            for h in handles:
                self.assertTrue(isinstance(h, PollData))
        except TSS2_Exception as e:
            if e.rc != lib.TSS2_TCTI_RC_NOT_IMPLEMENTED:
                raise e
            else:
                self.skipTest(f"get_poll_handles not supported by {tcti_name}")

    def test_custom_pytcti_set_locality(self):
        t = MyTCTI(self.tcti)
        t.set_locality(TPMA_LOCALITY.TWO)

    def test_custom_pytcti_make_sticky(self):
        t = MyTCTI(None)
        t._error = None
        t.make_sticky(0, 0)
        t.make_sticky(0, 1)
        t.make_sticky(0, False)

        # Test that throwing an exception shows the originating exception
        t._error = RuntimeError("Bills Error")
        with self.assertRaises(RuntimeError, msg="Bills Error"):
            t.make_sticky(5, True)

        t._v2 = None
        with self.assertRaises(TSS2_Exception):
            t.make_sticky(0, 0)

    def test_custom_pytcti_version(self):
        t = MyTCTI(None)
        self.assertEqual(t.version, 2)

    def test_custom_pytcti_magic(self):
        t = MyTCTI(None)
        magic = b"PYTCTI\x00\x00"
        self.assertEqual(t.magic, magic)

        # max magic len
        magic = b"THISISIT"
        t = MyTCTI(None, magic)
        self.assertEqual(t.magic, magic)

        # small magic len
        magic = b"COOL"
        t = MyTCTI(None, magic)
        self.assertEqual(t.magic, magic)

        # min magic
        magic = b""
        t = MyTCTI(None, magic)
        self.assertEqual(t.magic, magic)

        with self.assertRaises(ValueError):
            MyTCTI(None, b"THISISTOOBIG")

    def test_custom_pytcti_ctx_manager_finalize(self):
        with MyTCTI(self.tcti) as t:
            e = ESAPI(t)
            r = e.get_random(4)
            self.assertEqual(len(r), 4)
            e.startup(TPM2_SU.CLEAR)

        self.assertTrue(t.is_finalized)

    def test_custom_pytcti_finalize_error(self):

        t = MyTCTI(self.tcti)
        t._error = RuntimeError("Bills Error 2")
        with self.assertRaises(RuntimeError, msg="Bills Error 2"):
            t.finalize()

    def test_is_available(self):
        self.assertTrue(TCTILdr.is_available())
        self.assertFalse(TCTILdr.is_available("this-tcti-doesnt-exist"))


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: BSD-2
import unittest
from io import BytesIO
from tpm2_pytss.command_parser import (
    read_command,
    read_response,
    read_command_header,
    UINT32,
    read_command_sessions,
    TPM_ENCRYPTED_PARAMETER,
    tpm2_command,
    tpm2_response,
)
from tpm2_pytss.constants import (
    TPM2_ST,
    TPM2_CC,
    TPM2_SU,
    TPM2_RC,
    TPMA_SESSION,
    TPM2_SE,
    TPM2_ALG,
    TSS2_RC,
)
from tpm2_pytss.types import (
    TPM2_HANDLE,
    TPM2B_NONCE,
    TPM2B_AUTH,
    TPM2B_ENCRYPTED_SECRET,
    TPMT_SYM_DEF,
    TPM2B_DIGEST,
    TPM2B_SENSITIVE_DATA,
)
from tpm2_pytss.TSS2_Exception import TSS2_Exception


class CommandParserTest(unittest.TestCase):
    def test_startup_command(self):
        buf = b""
        buf += TPM2_CC.Startup.marshal()
        buf += TPM2_SU.CLEAR.marshal()
        tag = TPM2_ST.NO_SESSIONS.marshal()
        # commandSize is size of tag, bytes for commandSize and size of parameters
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        command = read_command(fp)
        self.assertEqual(len(fp.getvalue()), fp.tell())
        self.assertEqual(command.command_code, TPM2_CC.Startup)
        self.assertEqual(len(command.handles), 0)
        self.assertEqual(len(command.sessions), 0)
        self.assertEqual(len(command.parameters), 1)
        self.assertEqual(command.parameters, (TPM2_SU.CLEAR,))

    def test_startup_response(self):
        buf = b""
        buf += TPM2_RC.SUCCESS.marshal()
        tag = TPM2_ST.NO_SESSIONS.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        response = read_response(fp, TPM2_CC.Startup)
        self.assertEqual(len(fp.getvalue()), fp.tell())
        self.assertEqual(response.response_code, TPM2_RC.SUCCESS)
        self.assertEqual(response.handle, None)
        self.assertEqual(len(response.sessions), 0)
        self.assertEqual(len(response.parameters), 0)

    def test_failure_response(self):
        buf = b""
        buf += TPM2_RC.FAILURE.marshal()
        tag = TPM2_ST.NO_SESSIONS.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        response = read_response(fp, TPM2_CC.Startup)
        self.assertEqual(len(fp.getvalue()), fp.tell())
        self.assertEqual(response.response_code, TPM2_RC.FAILURE)
        self.assertEqual(response.handle, None)
        self.assertEqual(len(response.sessions), 0)
        self.assertEqual(len(response.parameters), 0)

    def test_read_command_header(self):
        fp = BytesIO()
        fp.write(TPM2_ST.SESSIONS.marshal())
        fp.write(UINT32(10).marshal())
        fp.write(TPM2_CC.Startup.marshal())
        fp.seek(0)
        print(int(10).to_bytes(4, byteorder="big"))

        tag, cc, left = read_command_header(fp)

        self.assertEqual(tag, TPM2_ST.SESSIONS)
        self.assertEqual(cc, TPM2_CC.Startup)
        self.assertEqual(left, 0)

    def test_read_command_sessions(self):
        fp = BytesIO()
        for i in range(0, 3):
            fp.write(TPM2_HANDLE(i).marshal())
            fp.write(TPM2B_NONCE(32 * bytes([i])).marshal())
            fp.write(TPMA_SESSION(i).marshal())
            fp.write(TPM2B_AUTH(bytes([i] * 32)).marshal())

        size = UINT32(len(fp.getvalue()))
        newfp = BytesIO()
        newfp.write(size.marshal())
        newfp.write(fp.getvalue())
        newfp.seek(0)

        sessions = read_command_sessions(newfp)

        self.assertEqual(len(sessions), 3)
        self.assertEqual(newfp.tell(), len(newfp.getvalue()))

        for i in range(0, len(sessions)):
            self.assertEqual(sessions[i].handle, i)
            self.assertEqual(sessions[i].attributes, i)
            self.assertEqual(sessions[i].nonce, bytes([i] * 32))
            self.assertEqual(sessions[i].authorization, bytes([i] * 32))

    def test_start_auth_session_command(self):
        buf = b""
        buf += TPM2_CC.StartAuthSession.marshal()
        buf += TPM2_HANDLE(1).marshal()
        buf += TPM2_HANDLE(2).marshal()
        # session here
        session_buf = b""
        session_buf += TPM2_HANDLE(1234).marshal()
        session_buf += TPM2B_NONCE(b"1234").marshal()
        session_buf += TPMA_SESSION(234).marshal()
        session_buf += TPM2B_AUTH(b"1234").marshal()
        session_size = UINT32(len(session_buf)).marshal()
        buf += session_size
        buf += session_buf
        # parameters here
        buf += TPM2B_NONCE(b"3").marshal()
        buf += TPM2B_ENCRYPTED_SECRET(b"4").marshal()
        buf += TPM2_SE.HMAC.marshal()
        buf += TPMT_SYM_DEF.parse("aes128cbc").marshal()
        buf += TPM2_ALG.SHA256.marshal()
        tag = TPM2_ST.SESSIONS.marshal()
        # commandSize is size of tag, bytes for commandSize and size of parameters
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        command = read_command(fp)

        self.assertEqual(command.tag, TPM2_ST.SESSIONS)
        self.assertEqual(len(command.handles), 2)
        self.assertEqual(command.handles[0], 1)
        self.assertEqual(command.handles[1], 2)

        self.assertEqual(len(command.sessions), 1)
        self.assertEqual(command.sessions[0].handle, TPM2_HANDLE(1234))
        self.assertEqual(command.sessions[0].nonce, TPM2B_NONCE(b"1234"))
        self.assertEqual(command.sessions[0].attributes, TPMA_SESSION(234))
        self.assertEqual(command.sessions[0].authorization, TPM2B_AUTH(b"1234"))

        self.assertEqual(len(command.parameters), 5)
        self.assertEqual(command.parameters[0], TPM2B_NONCE(b"3"))
        self.assertEqual(command.parameters[1], TPM2B_ENCRYPTED_SECRET(b"4"))
        self.assertEqual(command.parameters[2], TPM2_SE.HMAC)
        self.assertEqual(command.parameters[3].algorithm, TPM2_ALG.AES)
        self.assertEqual(command.parameters[3].mode.sym, TPM2_ALG.CBC)
        self.assertEqual(command.parameters[3].keyBits.sym, 128)
        self.assertEqual(command.parameters[4], TPM2_ALG.SHA256)

    def test_start_auth_session_response(self):
        buf = b""
        buf += TPM2_RC.SUCCESS.marshal()
        # handle here
        buf += TPM2_HANDLE(1).marshal()
        # parameters here
        parameter_buf = TPM2B_NONCE(b"falafel").marshal()
        parameter_size = UINT32(len(parameter_buf)).marshal()
        buf += parameter_size + parameter_buf
        # sessions here
        session_buf = TPM2B_NONCE(b"1234").marshal()
        session_buf += TPMA_SESSION(123).marshal()
        session_buf += TPM2B_DIGEST(b"1234").marshal()
        session_size = UINT32(len(session_buf)).marshal()
        buf += session_size + session_buf
        tag = TPM2_ST.SESSIONS.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        response = read_response(fp, TPM2_CC.StartAuthSession)

        self.assertEqual(response.tag, TPM2_ST.SESSIONS)
        self.assertEqual(response.response_code, TPM2_RC.SUCCESS)
        self.assertEqual(response.handle, TPM2_HANDLE(1))

        self.assertEqual(len(response.parameters), 1)
        self.assertEqual(response.parameters[0], TPM2B_NONCE(b"falafel"))

        self.assertEqual(len(response.sessions), 1)
        self.assertEqual(response.sessions[0].nonce, TPM2B_NONCE(b"1234"))
        self.assertEqual(response.sessions[0].attributes, TPMA_SESSION(123))
        self.assertEqual(response.sessions[0].acknowledgment, TPM2B_DIGEST(b"1234"))

    def test_encrypt_parameter(self):
        buf = b""
        buf += TPM2_RC.SUCCESS.marshal()
        parameter_buf = TPM2B_DIGEST(b"falafel").marshal()
        parameter_size = UINT32(len(parameter_buf)).marshal()
        buf += parameter_size + parameter_buf
        session_buf = TPM2B_NONCE(b"1234").marshal()
        session_buf += TPMA_SESSION.ENCRYPT.marshal()
        session_buf += TPM2B_DIGEST(b"1234").marshal()
        session_size = UINT32(len(session_buf)).marshal()
        buf += session_size + session_buf
        tag = TPM2_ST.SESSIONS.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        response = read_response(fp, TPM2_CC.GetRandom)

        self.assertEqual(response.tag, TPM2_ST.SESSIONS)
        self.assertEqual(response.response_code, TPM2_RC.SUCCESS)

        self.assertEqual(len(response.parameters), 1)
        self.assertIsInstance(response.parameters[0], TPM_ENCRYPTED_PARAMETER)
        self.assertEqual(response.parameters[0].underlying_type, TPM2B_DIGEST)
        self.assertEqual(response.parameters[0], b"falafel")

        self.assertEqual(len(response.sessions), 1)
        self.assertEqual(response.sessions[0].nonce, TPM2B_NONCE(b"1234"))
        self.assertEqual(response.sessions[0].attributes, TPMA_SESSION.ENCRYPT)
        self.assertEqual(response.sessions[0].acknowledgment, TPM2B_DIGEST(b"1234"))

    def test_decrypt_parameter(self):
        buf = b""
        buf += TPM2_CC.StirRandom.marshal()
        session_buf = b""
        session_buf += TPM2_HANDLE(1234).marshal()
        session_buf += TPM2B_NONCE(b"1234").marshal()
        session_buf += TPMA_SESSION.DECRYPT.marshal()
        session_buf += TPM2B_AUTH(b"1234").marshal()
        session_size = UINT32(len(session_buf)).marshal()
        buf += session_size + session_buf
        buf += TPM2B_SENSITIVE_DATA(b"falafel").marshal()
        tag = TPM2_ST.SESSIONS.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        command = read_command(fp)

        self.assertEqual(command.tag, TPM2_ST.SESSIONS)

        self.assertEqual(len(command.sessions), 1)
        self.assertEqual(command.sessions[0].handle, TPM2_HANDLE(1234))
        self.assertEqual(command.sessions[0].nonce, TPM2B_NONCE(b"1234"))
        self.assertEqual(command.sessions[0].attributes, TPMA_SESSION.DECRYPT)
        self.assertEqual(command.sessions[0].authorization, TPM2B_AUTH(b"1234"))

        self.assertIsInstance(command.parameters[0], TPM_ENCRYPTED_PARAMETER)
        self.assertEqual(command.parameters[0].underlying_type, TPM2B_SENSITIVE_DATA)
        self.assertEqual(command.parameters[0], b"falafel")

    def test_command_bad_tag(self):
        buf = b""
        buf += TPM2_CC.Startup.marshal()
        buf += TPM2_SU.CLEAR.marshal()
        tag = TPM2_ST.NULL.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)

        with self.assertRaises(TSS2_Exception) as e:
            read_command(fp)

        self.assertEqual(e.exception.rc, TPM2_RC.BAD_TAG)

    def test_response_bad_tag(self):
        buf = b""
        buf += TPM2_RC.SUCCESS.marshal()
        tag = TPM2_ST.ATTEST_NV.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)

        with self.assertRaises(TSS2_Exception) as e:
            read_response(fp, TPM2_CC.Startup)

        self.assertEqual(e.exception.rc, TPM2_RC.BAD_TAG)

    def test_tpm_encrypted_parameter(self):
        p = TPM_ENCRYPTED_PARAMETER(TPM2B_DIGEST, b"falafel")

        self.assertEqual(bytes(p), b"falafel")

        marshaled = p.marshal()
        self.assertEqual(marshaled, b"\x00\x07falafel")

        self.assertNotEqual(p, "falafel")

    def test_missing_commands(self):
        with self.assertRaises(TSS2_Exception) as e:
            tpm2_command.lookup_command_class(TPM2_CC.LAST + 1)
        self.assertEqual(e.exception.rc, TSS2_RC.BASE_RC_NOT_IMPLEMENTED)

        with self.assertRaises(TSS2_Exception) as e:
            tpm2_response.lookup_response_class(TPM2_CC.LAST + 1)
        self.assertEqual(e.exception.rc, TSS2_RC.BASE_RC_NOT_IMPLEMENTED)

    def test_without_encrypted_parameter(self):
        buf = b""
        buf += TPM2_RC.SUCCESS.marshal()
        parameter_buf = TPM2B_DIGEST(b"falafel").marshal()
        parameter_size = UINT32(len(parameter_buf)).marshal()
        buf += parameter_size + parameter_buf
        session_buf = TPM2B_NONCE(b"1234").marshal()
        session_buf += TPMA_SESSION(0).marshal()
        session_buf += TPM2B_DIGEST(b"1234").marshal()
        session_size = UINT32(len(session_buf)).marshal()
        buf += session_size + session_buf
        tag = TPM2_ST.SESSIONS.marshal()
        size = len(tag) + 4 + len(buf)
        size_bytes = size.to_bytes(4, byteorder="big")
        fp = BytesIO()
        fp.write(tag)
        fp.write(size_bytes)
        fp.write(buf)
        fp.seek(0)
        response = read_response(fp, TPM2_CC.GetRandom)

        self.assertIsInstance(response.parameters[0], TPM2B_DIGEST)



#!/usr/bin/python3 -u
"""
SPDX-License-Identifier: BSD-2
"""
import unittest

from tpm2_pytss import *
from tpm2_pytss.encoding import (
    base_encdec,
    json_encdec,
)


class SerializationTest(unittest.TestCase):
    def test_base_simple_tpm2b(self):
        enc = base_encdec()
        dig = TPM2B_DIGEST(b"falafel")
        ev = enc.encode(dig)
        self.assertEqual(ev, "66616c6166656c")

    def test_base_friendly_int(self):
        enc = base_encdec()
        ev = enc.encode(TPM2_ALG.SHA256)
        self.assertEqual(ev, 0x0B)

    def test_base_friendly_intlist(self):
        enc = base_encdec()
        ev = enc.encode(TPMA_OBJECT(TPMA_OBJECT.RESTRICTED | TPMA_OBJECT.FIXEDTPM))
        self.assertEqual(ev, 0x10002)

    def test_base_int(self):
        enc = base_encdec()
        ev = enc.encode(1337)
        self.assertEqual(ev, 1337)

    def test_base_complex_tpm2b(self):
        enc = base_encdec()
        p = TPM2B_ECC_POINT(TPMS_ECC_POINT(x=b"\x01", y="\x02"))
        ev = enc.encode(p)
        self.assertEqual(ev, {"x": "01", "y": "02"})

    def test_base_tpml(self):
        enc = base_encdec()
        al = TPML_ALG((TPM2_ALG.SHA256, TPM2_ALG.SHA384, TPM2_ALG.SHA512))
        ev = enc.encode(al)
        self.assertEqual(ev, [TPM2_ALG.SHA256, TPM2_ALG.SHA384, TPM2_ALG.SHA512])

    def test_base_TPMS_CAPABILITY_DATA(self):
        enc = base_encdec()
        algs = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.ALGS)
        algs.data.algorithms = TPML_ALG_PROPERTY(
            (TPMS_ALG_PROPERTY(alg=TPM2_ALG.SHA256, algProperties=TPMA_ALGORITHM.HASH),)
        )
        ev = enc.encode(algs)
        self.assertEqual(
            ev, {"capability": 0, "data": [{"alg": 11, "algProperties": 4}]}
        )

        handles = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.HANDLES)
        handles.data.handles = TPML_HANDLE((1,))
        ev = enc.encode(handles)
        self.assertEqual(ev, {"capability": 1, "data": [1]})

        commands = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.COMMANDS)
        cmd = TPM2_CC.NV_Write & TPMA_CC.NV & (2 << TPMA_CC.CHANDLES_SHIFT)
        commands.data.command = TPML_CCA((cmd,))
        ev = enc.encode(commands)
        self.assertEqual(ev, {"capability": 2, "data": [cmd]})

        ppcommands = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.PP_COMMANDS)
        ppcommands.data.ppCommands = TPML_CC((3,))
        ev = enc.encode(ppcommands)
        self.assertEqual(ev, {"capability": 3, "data": [3]})

        auditcommands = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.AUDIT_COMMANDS)
        auditcommands.data.auditCommands = TPML_CC((4,))
        ev = enc.encode(auditcommands)
        self.assertEqual(ev, {"capability": 4, "data": [4]})

        pcrs = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.PCRS)
        pcrsel = TPMS_PCR_SELECTION(
            hash=TPM2_ALG.SHA256, sizeofSelect=3, pcrSelect=b"\x81\x00\x00"
        )
        pcrs.data.assignedPCR = TPML_PCR_SELECTION((pcrsel,))
        ev = enc.encode(pcrs)
        self.assertEqual(
            ev, {"capability": 5, "data": [{"hash": 0x0B, "pcrSelect": [0, 7]}]}
        )

        tpm = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.TPM_PROPERTIES)
        prop = TPMS_TAGGED_PROPERTY(property=TPM2_PT_NV.BUFFER_MAX, value=0x300)
        tpm.data.tpmProperties = TPML_TAGGED_TPM_PROPERTY((prop,))
        ev = enc.encode(tpm)
        self.assertEqual(
            ev, {"capability": 6, "data": [{"property": 300, "value": 0x300}]}
        )

        pcrs = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.PCR_PROPERTIES)
        pprop = TPMS_TAGGED_PCR_SELECT(
            tag=TPM2_PT_PCR.DRTM_RESET, sizeofSelect=3, pcrSelect=b"\xFF\x00\x00"
        )
        pcrs.data.pcrProperties = TPML_TAGGED_PCR_PROPERTY((pprop,))
        ev = enc.encode(pcrs)
        self.assertEqual(
            ev,
            {
                "capability": 7,
                "data": [{"tag": 0x12, "pcrSelect": [0, 1, 2, 3, 4, 5, 6, 7]}],
            },
        )

        curves = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.ECC_CURVES)
        curves.data.eccCurves = TPML_ECC_CURVE((TPM2_ECC.SM2_P256,))
        ev = enc.encode(curves)
        self.assertEqual(ev, {"capability": 8, "data": [0x20]})

    def test_base_TPMS_ATTEST(self):
        enc = base_encdec()
        certify = TPMS_ATTEST(type=TPM2_ST.ATTEST_CERTIFY)
        certify.attested.certify = TPMS_CERTIFY_INFO(
            name=b"\x01", qualifiedName=b"\x02"
        )
        ev = enc.encode(certify)
        self.assertEqual(ev["type"], 0x8017)
        self.assertEqual(ev["attested"], {"name": "01", "qualifiedName": "02"})

        creation = TPMS_ATTEST(type=TPM2_ST.ATTEST_CREATION)
        creation.attested.creation = TPMS_CREATION_INFO(
            objectName=b"\x01", creationHash=b"\x02"
        )
        ev = enc.encode(creation)
        self.assertEqual(ev["type"], 0x801A)
        self.assertEqual(ev["attested"], {"objectName": "01", "creationHash": "02"})

        quote = TPMS_ATTEST(type=TPM2_ST.ATTEST_QUOTE)
        sel = TPMS_PCR_SELECTION(
            hash=TPM2_ALG.SHA256, sizeofSelect=3, pcrSelect=b"\x00\x81\x00"
        )
        quote.attested.quote = TPMS_QUOTE_INFO(
            pcrSelect=TPML_PCR_SELECTION((sel,)), pcrDigest=b"\x01" * 32
        )
        ev = enc.encode(quote)
        self.assertEqual(ev["type"], 0x8018)
        self.assertEqual(
            ev["attested"],
            {
                "pcrSelect": [{"hash": TPM2_ALG.SHA256, "pcrSelect": [8, 15]}],
                "pcrDigest": "01" * 32,
            },
        )

        command = TPMS_ATTEST(type=TPM2_ST.ATTEST_COMMAND_AUDIT)
        command.attested.commandAudit = TPMS_COMMAND_AUDIT_INFO(
            auditCounter=1337,
            digestAlg=TPM2_ALG.SHA256,
            auditDigest=b"\x01",
            commandDigest=b"\x02",
        )
        ev = enc.encode(command)
        self.assertEqual(ev["type"], TPM2_ST.ATTEST_COMMAND_AUDIT)
        self.assertEqual(
            ev["attested"],
            {
                "auditCounter": 1337,
                "digestAlg": TPM2_ALG.SHA256,
                "auditDigest": "01",
                "commandDigest": "02",
            },
        )

        session = TPMS_ATTEST(type=TPM2_ST.ATTEST_SESSION_AUDIT)
        session.attested.sessionAudit = TPMS_SESSION_AUDIT_INFO(
            exclusiveSession=True, sessionDigest=b"\x01"
        )
        ev = enc.encode(session)
        self.assertEqual(ev["type"], TPM2_ST.ATTEST_SESSION_AUDIT)
        self.assertEqual(ev["attested"], {"exclusiveSession": 1, "sessionDigest": "01"})

        time = TPMS_ATTEST(type=TPM2_ST.ATTEST_TIME)
        time.attested.time = TPMS_TIME_ATTEST_INFO(
            time=TPMS_TIME_INFO(
                time=1234,
                clockInfo=TPMS_CLOCK_INFO(
                    clock=1024, resetCount=2, restartCount=3, safe=False
                ),
            ),
            firmwareVersion=1337,
        )
        ev = enc.encode(time)
        self.assertEqual(ev["type"], TPM2_ST.ATTEST_TIME)
        self.assertEqual(
            ev["attested"],
            {
                "firmwareVersion": 1337,
                "time": {
                    "time": 1234,
                    "clockInfo": {
                        "clock": 1024,
                        "resetCount": 2,
                        "restartCount": 3,
                        "safe": 0,
                    },
                },
            },
        )

        nv = TPMS_ATTEST(type=TPM2_ST.ATTEST_NV)
        nv.attested.nv = TPMS_NV_CERTIFY_INFO(
            indexName=b"\x01", offset=2, nvContents=b"\x03"
        )
        ev = enc.encode(nv)
        self.assertEqual(ev["type"], TPM2_ST.ATTEST_NV)
        self.assertEqual(
            ev["attested"], {"indexName": "01", "offset": 2, "nvContents": "03"}
        )

    def test_base_TPMT_SYM_DEF(self):
        enc = base_encdec()
        sym = TPMT_SYM_DEF(algorithm=TPM2_ALG.AES)
        sym.keyBits.aes = 128
        sym.mode.aes = TPM2_ALG.CFB
        ev = enc.encode(sym)
        self.assertEqual(
            ev, {"algorithm": TPM2_ALG.AES, "keyBits": 128, "mode": TPM2_ALG.CFB}
        )

        xor = TPMT_SYM_DEF(algorithm=TPM2_ALG.XOR)
        xor.keyBits.exclusiveOr = TPM2_ALG.SHA1
        ev = enc.encode(xor)
        self.assertEqual(ev, {"algorithm": TPM2_ALG.XOR, "keyBits": TPM2_ALG.SHA1})

    def test_base_TPMT_KEYEDHASH_SCHEME(self):
        enc = base_encdec()
        hmac = TPMT_KEYEDHASH_SCHEME(scheme=TPM2_ALG.HMAC)
        hmac.details.hmac.hashAlg = TPM2_ALG.SHA256
        ev = enc.encode(hmac)
        self.assertEqual(
            ev, {"scheme": TPM2_ALG.HMAC, "details": {"hashAlg": TPM2_ALG.SHA256}}
        )

        xor = TPMT_KEYEDHASH_SCHEME(scheme=TPM2_ALG.XOR)
        xor.details.exclusiveOr = TPMS_SCHEME_XOR(
            hashAlg=TPM2_ALG.SHA256, kdf=TPM2_ALG.KDF2
        )
        ev = enc.encode(xor)
        self.assertEqual(
            ev,
            {
                "scheme": TPM2_ALG.XOR,
                "details": {"hashAlg": TPM2_ALG.SHA256, "kdf": TPM2_ALG.KDF2},
            },
        )

    def test_base_TPMT_SIG_SCHEME(self):
        enc = base_encdec()
        ecdaa = TPMT_SIG_SCHEME(scheme=TPM2_ALG.ECDAA)
        ecdaa.details.ecdaa = TPMS_SCHEME_ECDAA(hashAlg=TPM2_ALG.SHA256, count=2)
        ev = enc.encode(ecdaa)
        self.assertEqual(
            ev,
            {
                "scheme": TPM2_ALG.ECDAA,
                "details": {"hashAlg": TPM2_ALG.SHA256, "count": 2},
            },
        )

        sig = TPMT_SIG_SCHEME(scheme=TPM2_ALG.ECDSA)
        sig.details.any.hashAlg = TPM2_ALG.SHA256
        ev = enc.encode(sig)
        self.assertEqual(
            ev, {"scheme": TPM2_ALG.ECDSA, "details": {"hashAlg": TPM2_ALG.SHA256}}
        )

    def test_base_TPMT_KDF_SCHEME(self):
        enc = base_encdec()
        kdf = TPMT_KDF_SCHEME(scheme=TPM2_ALG.KDF2)
        kdf.details.mgf1.hashAlg = TPM2_ALG.SHA256
        ev = enc.encode(kdf)
        self.assertEqual(
            ev, {"scheme": TPM2_ALG.KDF2, "details": {"hashAlg": TPM2_ALG.SHA256}}
        )

    def test_base_TPMT_ASYM_SCHEME(self):
        enc = base_encdec()
        ecdaa = TPMT_ASYM_SCHEME(scheme=TPM2_ALG.ECDAA)
        ecdaa.details.ecdaa = TPMS_SCHEME_ECDAA(hashAlg=TPM2_ALG.SHA256, count=2)
        ev = enc.encode(ecdaa)
        self.assertEqual(
            ev,
            {
                "scheme": TPM2_ALG.ECDAA,
                "details": {"count": 2, "hashAlg": TPM2_ALG.SHA256},
            },
        )

        rsaes = TPMT_ASYM_SCHEME(scheme=TPM2_ALG.RSAES)
        ev = enc.encode(rsaes)
        self.assertEqual(ev, {"scheme": TPM2_ALG.RSAES})

        scheme = TPMT_ASYM_SCHEME(scheme=TPM2_ALG.RSASSA)
        scheme.details.rsassa.hashAlg = TPM2_ALG.SHA256
        ev = enc.encode(scheme)
        self.assertEqual(
            ev, {"scheme": TPM2_ALG.RSASSA, "details": {"hashAlg": TPM2_ALG.SHA256}}
        )

    def test_base_TPMT_SINGATURE(self):
        enc = base_encdec()
        rsa = TPMT_SIGNATURE(sigAlg=TPM2_ALG.RSASSA)
        rsa.signature.rsassa = TPMS_SIGNATURE_RSA(hash=TPM2_ALG.SHA256, sig=b"\x01")
        ev = enc.encode(rsa)
        self.assertEqual(
            ev,
            {
                "sigAlg": TPM2_ALG.RSASSA,
                "signature": {"hash": TPM2_ALG.SHA256, "sig": "01"},
            },
        )

        ecc = TPMT_SIGNATURE(sigAlg=TPM2_ALG.ECDSA)
        ecc.signature.ecdsa = TPMS_SIGNATURE_ECC(
            hash=TPM2_ALG.SHA256, signatureR=b"\x02", signatureS=b"\x03"
        )
        ev = enc.encode(ecc)
        self.assertEqual(
            ev,
            {
                "sigAlg": TPM2_ALG.ECDSA,
                "signature": {
                    "hash": TPM2_ALG.SHA256,
                    "signatureR": "02",
                    "signatureS": "03",
                },
            },
        )

        hmac = TPMT_SIGNATURE(sigAlg=TPM2_ALG.HMAC)
        hmac.signature.hmac.hashAlg = TPM2_ALG.SHA256
        hmac.signature.hmac.digest.sha256 = b"\x01" * 32
        ev = enc.encode(hmac)
        self.assertEqual(
            ev,
            {
                "sigAlg": TPM2_ALG.HMAC,
                "signature": {"hashAlg": TPM2_ALG.SHA256, "digest": "01" * 32},
            },
        )

    def test_base_TPMT_PUBLIC_PARMS(self):
        enc = base_encdec()
        keyedhash = TPMT_PUBLIC_PARMS(type=TPM2_ALG.KEYEDHASH)
        keyedhash.parameters.keyedHashDetail.scheme = TPMT_KEYEDHASH_SCHEME(
            scheme=TPM2_ALG.HMAC
        )
        keyedhash.parameters.keyedHashDetail.scheme.details.hmac.hashAlg = (
            TPM2_ALG.SHA256
        )
        ev = enc.encode(keyedhash)
        self.assertEqual(ev["type"], TPM2_ALG.KEYEDHASH)
        self.assertEqual(
            ev["parameters"],
            {
                "scheme": {
                    "details": {"hashAlg": TPM2_ALG.SHA256},
                    "scheme": TPM2_ALG.HMAC,
                },
            },
        )

        sym = TPMT_PUBLIC_PARMS(type=TPM2_ALG.SYMCIPHER)
        sym.parameters.symDetail.sym = TPMT_SYM_DEF_OBJECT(algorithm=TPM2_ALG.AES)
        sym.parameters.symDetail.sym.keyBits.aes = 128
        sym.parameters.symDetail.sym.mode.aes = TPM2_ALG.CFB
        ev = enc.encode(sym)
        self.assertEqual(ev["type"], TPM2_ALG.SYMCIPHER)
        self.assertEqual(
            ev["parameters"],
            {"sym": {"algorithm": TPM2_ALG.AES, "keyBits": 128, "mode": TPM2_ALG.CFB}},
        )

        rsa = TPMT_PUBLIC_PARMS(type=TPM2_ALG.RSA)
        parms = TPMS_RSA_PARMS(keyBits=2048)
        parms.symmetric.algorithm = TPM2_ALG.NULL
        parms.scheme.scheme = TPM2_ALG.NULL
        rsa.parameters.rsaDetail = parms
        ev = enc.encode(rsa)
        self.assertEqual(ev["type"], TPM2_ALG.RSA)
        self.assertEqual(
            ev["parameters"],
            {
                "exponent": 0,
                "keyBits": 2048,
                "scheme": {"scheme": TPM2_ALG.NULL},
                "symmetric": {"algorithm": TPM2_ALG.NULL},
            },
        )

        ecc = TPMT_PUBLIC_PARMS(type=TPM2_ALG.ECC)
        parms = TPMS_ECC_PARMS(curveID=TPM2_ALG.SM2)
        parms.symmetric.algorithm = TPM2_ALG.NULL
        parms.scheme.scheme = TPM2_ALG.NULL
        parms.kdf.scheme = TPM2_ALG.NULL
        ecc.parameters.eccDetail = parms
        ev = enc.encode(ecc)
        self.assertEqual(ev["type"], TPM2_ALG.ECC)
        self.assertEqual(
            ev["parameters"],
            {
                "curveID": TPM2_ALG.SM2,
                "kdf": {"scheme": TPM2_ALG.NULL},
                "scheme": {"scheme": TPM2_ALG.NULL},
                "symmetric": {"algorithm": TPM2_ALG.NULL},
            },
        )

    def test_base_TPMT_PUBLIC(self):
        enc = base_encdec()
        keyedhash = TPMT_PUBLIC(type=TPM2_ALG.KEYEDHASH)
        keyedhash.unique.keyedHash = b"\x22" * 32
        keyedhash.parameters.keyedHashDetail.scheme.scheme = TPM2_ALG.NULL
        ev = enc.encode(keyedhash)
        self.assertEqual(ev["type"], TPM2_ALG.KEYEDHASH)
        self.assertEqual(ev["unique"], "22" * 32)

        rsa = TPMT_PUBLIC(type=TPM2_ALG.RSA)
        rsa.unique.rsa = b"\x33" * 32
        ev = enc.encode(rsa)
        self.assertEqual(ev["type"], TPM2_ALG.RSA)
        self.assertEqual(ev["unique"], "33" * 32)

        ecc = TPMT_PUBLIC(type=TPM2_ALG.ECC)
        ecc.unique.ecc.x = b"\x04"
        ecc.unique.ecc.y = b"\x05"
        ev = enc.encode(ecc)
        self.assertEqual(ev["type"], TPM2_ALG.ECC)
        self.assertEqual(ev["unique"], {"x": "04", "y": "05"})

    def test_base_TPMT_SENSITIVE(self):
        enc = base_encdec()
        rsa = TPMT_SENSITIVE(sensitiveType=TPM2_ALG.RSA)
        rsa.sensitive.rsa = b"\x11" * 32
        ev = enc.encode(rsa)
        self.assertEqual(ev["sensitiveType"], TPM2_ALG.RSA)
        self.assertEqual(ev["sensitive"], "11" * 32)

    def test_base_bad(self):
        enc = base_encdec()
        with self.assertRaises(TypeError) as e:
            enc.encode(TPMU_KDF_SCHEME())
        self.assertEqual(str(e.exception), "tried to encode union TPMU_KDF_SCHEME")

        with self.assertRaises(ValueError) as e:
            null = TPMT_PUBLIC(type=TPM2_ALG.NULL)
            enc.encode(null)
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field parameters in TPMT_PUBLIC",
        )

        with self.assertRaises(TypeError) as e:
            enc.encode(list())
        self.assertEqual(str(e.exception), "unable to encode value of type list")

    def test_base_decode_friendly_intlist(self):
        dec = base_encdec()
        attrs = int(TPMA_OBJECT.USERWITHAUTH | TPMA_OBJECT.DECRYPT)
        dv = dec.decode(TPMA_OBJECT(), attrs)
        self.assertEqual(dv, TPMA_OBJECT.USERWITHAUTH | TPMA_OBJECT.DECRYPT)
        self.assertIsInstance(dv, TPMA_OBJECT)

    def test_base_decode_friendly_int(self):
        dec = base_encdec()
        alg = int(TPM2_ALG.SHA256)
        dv = dec.decode(TPM2_ALG(), alg)
        self.assertEqual(dv, TPM2_ALG.SHA256)
        self.assertIsInstance(dv, TPM2_ALG)

    def test_base_decode_int(self):
        dec = base_encdec()
        i = 1234
        dv = dec.decode(int(), i)
        self.assertEqual(dv, 1234)
        self.assertIsInstance(dv, int)

    def test_base_decode_simple_tpm2b(self):
        dec = base_encdec()
        hstr = "01020304"
        dv = dec.decode(TPM2B_DIGEST(), hstr)
        self.assertEqual(dv.buffer, b"\x01\x02\x03\x04")
        self.assertEqual(dv.size, 4)
        self.assertIsInstance(dv, TPM2B_DIGEST)

    def test_base_decode_complex_tpm2b(self):
        dec = base_encdec()
        points = {"x": "01", "y": "02"}
        dv = dec.decode(TPM2B_ECC_POINT(), points)
        self.assertEqual(dv.point.x, b"\x01")
        self.assertEqual(dv.point.y, b"\x02")
        self.assertIsInstance(dv, TPM2B_ECC_POINT)

    def test_base_decode_tpml(self):
        dec = base_encdec()
        cmds = (int(TPM2_CC.Create), int(TPM2_CC.CreatePrimary))
        dv = dec.decode(TPML_CC(), cmds)
        self.assertEqual(len(dv), 2)
        self.assertIsInstance(dv, TPML_CC)
        self.assertEqual(dv[0], TPM2_CC.Create)
        self.assertEqual(dv[1], TPM2_CC.CreatePrimary)
        self.assertIsInstance(dv[0], TPM2_CC)
        self.assertIsInstance(dv[1], TPM2_CC)

    def test_base_decode_struct(self):
        dec = base_encdec()
        templ = {"type": TPM2_ALG.RSA, "parameters": {"exponent": 1234, "keyBits": 1}}
        dv = dec.decode(TPMT_PUBLIC_PARMS(), templ)
        self.assertIsInstance(dv, TPMT_PUBLIC_PARMS)
        self.assertEqual(dv.type, TPM2_ALG.RSA)
        self.assertEqual(dv.parameters.rsaDetail.exponent, 1234)
        self.assertEqual(dv.parameters.rsaDetail.keyBits, 1)

    def test_base_decode_pcrselect(self):
        dec = base_encdec()
        sel = {"pcrSelect": [8, 15]}
        dv = dec.decode(TPMS_PCR_SELECT(), sel)
        self.assertEqual(dv.sizeofSelect, 2)
        self.assertEqual(bytes(dv.pcrSelect), b"\x00\x81\x00\x00")

    def test_base_decode_strict(self):
        points = {"x": "01", "y": "02", "z": "03"}
        decoder = base_encdec(strict=True)
        with self.assertRaises(ValueError) as e:
            decoder.decode(TPMS_ECC_POINT(), points)
        self.assertEqual(str(e.exception), "unknown field(s) z in source")

    def test_base_decode_case_insensitive(self):
        points = {"X": "01", "Y": "02"}
        dec = base_encdec(case_insensitive=True)
        dv = dec.decode(TPMS_ECC_POINT(), points)
        self.assertEqual(dv.x, b"\x01")
        self.assertEqual(dv.y, b"\x02")

    def test_base_decode_bad(self):
        dec = base_encdec()
        with self.assertRaises(TypeError) as e:
            dec.decode(TPMU_KDF_SCHEME(), {})
        self.assertEqual(str(e.exception), "tried to decode union TPMU_KDF_SCHEME")

        with self.assertRaises(TypeError) as e:
            dec.decode("", {})
        self.assertEqual(str(e.exception), "unable to decode value of type str")

    def test_base_bad_selector(self):
        enc = base_encdec()
        cap = TPMS_CAPABILITY_DATA(capability=TPM2_CAP.VENDOR_PROPERTY + 1)
        with self.assertRaises(ValueError) as e:
            enc.encode(cap)
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field data in TPMS_CAPABILITY_DATA",
        )

        att = TPMS_ATTEST(type=TPM2_ST.FU_MANIFEST + 1)
        with self.assertRaises(ValueError) as e:
            enc.encode(att)
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field attested in TPMS_ATTEST",
        )

        keyed = TPMT_KEYEDHASH_SCHEME(scheme=TPM2_ALG.LAST + 1)
        with self.assertRaises(ValueError) as e:
            enc.encode(keyed)
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field details in TPMT_KEYEDHASH_SCHEME",
        )

        sig = TPMT_SIGNATURE(sigAlg=TPM2_ALG.LAST + 1)
        with self.assertRaises(ValueError) as e:
            enc.encode(sig)
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field signature in TPMT_SIGNATURE",
        )

        dec = base_encdec()
        with self.assertRaises(ValueError) as e:
            dec._get_by_selector(TPMT_PUBLIC(), "badfield")
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field badfield in TPMT_PUBLIC",
        )

        with self.assertRaises(ValueError) as e:
            dec._get_by_selector(TPMT_PUBLIC(type=TPM2_ALG.NULL), "unique")
        self.assertEqual(
            str(e.exception),
            "unable to find union selector for field unique in TPMT_PUBLIC",
        )

    def test_json_enc_int(self):
        enc = json_encdec()
        ev = enc.encode(1234)
        self.assertEqual(ev, 1234)

        ev = enc.encode(0x100000002)
        self.assertEqual(ev, [1, 2])

    def test_json_enc_friendly_int(self):
        enc = json_encdec()
        ev = enc.encode(TPM2_RH.OWNER)
        self.assertEqual(ev, "owner")

        ev = enc.encode(TPM2_ALG.LAST + 1)
        self.assertIsInstance(ev, int)
        self.assertEqual(ev, int(TPM2_ALG.LAST + 1))

        ev = enc.encode(TPM2_GENERATED.VALUE)
        self.assertEqual(ev, "value")

    def test_json_enc_friendly_intlist(self):
        enc = json_encdec()
        ev = enc.encode(TPMA_OBJECT.USERWITHAUTH | TPMA_OBJECT.RESTRICTED)
        self.assertEqual(ev, {"userwithauth": 1, "restricted": 1})

        ev = enc.encode(TPMA_NV.AUTHREAD | TPM2_NT.COUNTER << TPMA_NV.TPM2_NT_SHIFT)
        self.assertEqual(ev, {"authread": 1, "nt": "counter"})

        ev = enc.encode(TPMA_CC.NV | 5 << TPMA_CC.CHANDLES_SHIFT | 1234)
        self.assertEqual(ev, {"nv": 1, "chandles": 5, "commandindex": 1234})

        ev = enc.encode(
            TPMA_LOCALITY.ZERO | TPMA_LOCALITY(5 << TPMA_LOCALITY.EXTENDED_SHIFT)
        )
        self.assertEqual(ev, {"zero": 1, "extended": 5})

    def test_json_dec_int(self):
        dec = json_encdec()
        dv = dec.decode(int(), 1234)
        self.assertEqual(dv, 1234)

        dv = dec.decode(int(), "0xAA")
        self.assertEqual(dv, 0xAA)

        dv = dec.decode(int(), [1, 2])
        self.assertEqual(dv, 0x100000002)

    def test_json_dec_friendly_int(self):
        dec = json_encdec()
        dv = dec.decode(TPM2_ALG(), "sha1")
        self.assertEqual(dv, TPM2_ALG.SHA1)

        dv = dec.decode(TPM2_ALG(), "0x4")
        self.assertEqual(dv, TPM2_ALG.SHA1)

        dv = dec.decode(TPM2_ALG(), int(TPM2_ALG.SHA1))
        self.assertEqual(dv, TPM2_ALG.SHA1)

    def test_json_dec_friendly_intlist(self):
        dec = json_encdec()
        dv = dec.decode(
            TPMA_OBJECT.RESTRICTED,
            {"userwithauth": 1, "restricted": 0, "sign_encrypt": 1},
        )
        self.assertEqual(dv, TPMA_OBJECT.USERWITHAUTH | TPMA_OBJECT.SIGN_ENCRYPT)

        dv = dec.decode(TPMA_OBJECT(), ["TPMA_OBJECT_RESTRICTED", "ADMINWITHPOLICY"])
        self.assertEqual(dv, TPMA_OBJECT.RESTRICTED | TPMA_OBJECT.ADMINWITHPOLICY)

        dv = dec.decode(
            TPMA_OBJECT(), f"0x{TPMA_OBJECT.USERWITHAUTH | TPMA_OBJECT.SIGN_ENCRYPT:x}"
        )
        self.assertEqual(dv, TPMA_OBJECT.USERWITHAUTH | TPMA_OBJECT.SIGN_ENCRYPT)

        dv = dec.decode(TPMA_NV(), {"written": "set", "nt": "counter"})
        self.assertEqual(dv, TPMA_NV.WRITTEN | TPM2_NT.COUNTER << TPMA_NV.TPM2_NT_SHIFT)

        dv = dec.decode(TPMA_CC(), {"v": 1, "commandindex": 1234, "chandles": 5})
        self.assertEqual(dv, TPMA_CC.V | 1234 | 5 << TPMA_CC.CHANDLES_SHIFT)

        dv = dec.decode(TPMA_LOCALITY(), {"one": 1, "extended": 5})
        self.assertEqual(dv, TPMA_LOCALITY(0xA0) | TPMA_LOCALITY.ONE)

    def test_json_dec_simple_tpm2b(self):
        dec = json_encdec()
        dv = dec.decode(TPM2B_DIGEST(), "01020304")
        self.assertEqual(dv, b"\x01\x02\x03\x04")

        dv = dec.decode(TPM2B_DIGEST(), "0xff")
        self.assertEqual(dv, b"\xFF")

        dv = dec.decode(TPM2B_DIGEST(), [1, 2, 3, 4])
        self.assertEqual(dv, b"\x01\x02\x03\x04")


if __name__ == "__main__":
    unittest.main()



from binascii import hexlify, unhexlify
from typing import Any, Union, List, Dict
from ._libtpm2_pytss import ffi
from .constants import (
    TPM_FRIENDLY_INT,
    TPMA_FRIENDLY_INTLIST,
    TPM2_CAP,
    TPM2_ST,
    TPM2_ALG,
    TPMA_NV,
    TPMA_CC,
    TPM2_NT,
    TPMA_LOCALITY,
)
from .types import (
    TPM_OBJECT,
    TPM2B_SIMPLE_OBJECT,
    TPML_OBJECT,
    TPM2B_PUBLIC,
    TPM2B_SENSITIVE,
    TPMT_PUBLIC,
    TPMT_SENSITIVE,
    TPMT_KDF_SCHEME,
    TPMT_ASYM_SCHEME,
    TPMT_RSA_SCHEME,
    TPMT_ECC_SCHEME,
    TPMT_SYM_DEF_OBJECT,
    TPMT_KEYEDHASH_SCHEME,
    TPMT_HA,
    TPMS_CAPABILITY_DATA,
    TPMS_ATTEST,
    TPMT_SIGNATURE,
    TPM2B_SENSITIVE_CREATE,
    TPM2B_ECC_POINT,
    TPM2B_NV_PUBLIC,
    TPM2B_CREATION_DATA,
    TPMT_SYM_DEF,
    TPMT_SIG_SCHEME,
    TPMT_RSA_DECRYPT,
    TPMT_PUBLIC_PARMS,
    TPM_BASE_OBJECT,
)


class base_encdec(object):
    """Base encoder/decoder for TPM types

    Args:
        strict (bool): If a exception should be raised for unknown fields during decoding, defaults to False.
        case_insensitive (bool): If field names should be case insensitive during decoding, defaults to False.
    """

    def __init__(self, strict: bool = False, case_insensitive: bool = False):
        self._strict = strict
        self._case_insensitive = case_insensitive

    def _is_union(self, val):
        if not hasattr(val, "_cdata"):
            return False
        to = ffi.typeof(val._cdata)
        if to.kind == "pointer":
            to = to.item
        if to.kind == "union":
            return True
        return False

    def _get_complex_field(self, val):
        for fn in dir(val._cdata):
            if fn != "size":
                return getattr(val, fn)
        return None

    def _set_complex_field(self, dst, f):
        for fn in dir(dst._cdata):
            if fn != "size":
                setattr(dst, fn, f)
                return
        raise ValueError(f"no complex field found for {dst.__class__.__name__}")

    def _get_element_type(self, dst):
        return type(dst[0])

    def _get_by_selector(self, val, field):
        if isinstance(val, TPMS_CAPABILITY_DATA):
            if val.capability == TPM2_CAP.ALGS:
                return val.data.algorithms
            if val.capability == TPM2_CAP.HANDLES:
                return val.data.handles
            if val.capability == TPM2_CAP.COMMANDS:
                return val.data.command
            if val.capability == TPM2_CAP.PP_COMMANDS:
                return val.data.ppCommands
            if val.capability == TPM2_CAP.AUDIT_COMMANDS:
                return val.data.auditCommands
            if val.capability == TPM2_CAP.PCRS:
                return val.data.assignedPCR
            if val.capability == TPM2_CAP.TPM_PROPERTIES:
                return val.data.tpmProperties
            if val.capability == TPM2_CAP.PCR_PROPERTIES:
                return val.data.pcrProperties
            if val.capability == TPM2_CAP.ECC_CURVES:
                return val.data.eccCurves
        elif isinstance(val, TPMS_ATTEST):
            if val.type == TPM2_ST.ATTEST_CERTIFY:
                return val.attested.certify
            if val.type == TPM2_ST.ATTEST_CREATION:
                return val.attested.creation
            if val.type == TPM2_ST.ATTEST_QUOTE:
                return val.attested.quote
            if val.type == TPM2_ST.ATTEST_COMMAND_AUDIT:
                return val.attested.commandAudit
            if val.type == TPM2_ST.ATTEST_SESSION_AUDIT:
                return val.attested.sessionAudit
            if val.type == TPM2_ST.ATTEST_TIME:
                return val.attested.time
            if val.type == TPM2_ST.ATTEST_NV:
                return val.attested.nv
        elif (
            isinstance(val, (TPMT_SYM_DEF, TPMT_SYM_DEF_OBJECT)) and field == "keyBits"
        ):
            if val.algorithm == TPM2_ALG.XOR:
                return val.keyBits.exclusiveOr
            elif val.algorithm in (TPM2_ALG.NULL, TPM2_ALG.ERROR):
                return None
            return val.keyBits.sym
        elif isinstance(val, (TPMT_SYM_DEF, TPMT_SYM_DEF_OBJECT)) and field == "mode":
            if val.algorithm in (TPM2_ALG.XOR, TPM2_ALG.NULL, TPM2_ALG.ERROR):
                return None
            return val.mode.sym
        elif isinstance(val, TPMT_KEYEDHASH_SCHEME):
            if val.scheme == TPM2_ALG.HMAC:
                return val.details.hmac
            if val.scheme == TPM2_ALG.XOR:
                return val.details.exclusiveOr
            if val.scheme == TPM2_ALG.NULL:
                return None
        elif isinstance(val, TPMT_SIG_SCHEME):
            if val.scheme == TPM2_ALG.ECDAA:
                return val.details.ecdaa
            return val.details.any
        elif isinstance(val, TPMT_KDF_SCHEME):
            if val.scheme in (TPM2_ALG.NULL, TPM2_ALG.ERROR):
                return None
            return val.details.mgf1
        elif isinstance(
            val, (TPMT_ASYM_SCHEME, TPMT_RSA_SCHEME, TPMT_RSA_DECRYPT, TPMT_ECC_SCHEME)
        ):
            if val.scheme == TPM2_ALG.ECDAA:
                return val.details.ecdaa
            if val.scheme in (TPM2_ALG.RSAES, TPM2_ALG.NULL, TPM2_ALG.ERROR):
                return None
            return val.details.anySig
        elif isinstance(val, TPMT_SIGNATURE):
            if val.sigAlg in (TPM2_ALG.RSASSA, TPM2_ALG.RSAPSS):
                return val.signature.rsassa
            if val.sigAlg in (
                TPM2_ALG.ECDSA,
                TPM2_ALG.ECDAA,
                TPM2_ALG.SM2,
                TPM2_ALG.ECSCHNORR,
            ):
                return val.signature.ecdsa
            if val.sigAlg == TPM2_ALG.HMAC:
                return val.signature.hmac
        elif (
            isinstance(val, (TPMT_PUBLIC_PARMS, TPMT_PUBLIC)) and field == "parameters"
        ):
            if val.type == TPM2_ALG.KEYEDHASH:
                return val.parameters.keyedHashDetail
            if val.type == TPM2_ALG.SYMCIPHER:
                return val.parameters.symDetail
            if val.type == TPM2_ALG.RSA:
                return val.parameters.rsaDetail
            if val.type == TPM2_ALG.ECC:
                return val.parameters.eccDetail
        elif isinstance(val, TPMT_PUBLIC) and field == "unique":
            if val.type in (TPM2_ALG.KEYEDHASH, TPM2_ALG.SYMCIPHER):
                return val.unique.keyedHash
            if val.type == TPM2_ALG.RSA:
                return val.unique.rsa
            if val.type == TPM2_ALG.ECC:
                return val.unique.ecc
        elif isinstance(val, TPMT_SENSITIVE):
            return val.sensitive.any
        elif isinstance(val, TPMT_HA):
            return bytes(val)
        raise ValueError(
            f"unable to find union selector for field {field} in {val.__class__.__name__}"
        )

    def encode(
        self,
        val: Union[
            TPMA_FRIENDLY_INTLIST,
            TPM_FRIENDLY_INT,
            int,
            TPM2B_SIMPLE_OBJECT,
            TPML_OBJECT,
            TPM_OBJECT,
            bytes,
        ],
    ) -> Union[int, str, Dict[str, Any], list, None]:
        """Encode a TPM type

        Args:
            val (Union[TPMA_FRIENDLY_INTLIST, TPM_FRIENDLY_INT, int, TPM2B_SIMPLE_OBJECT, TPML_OBJECT, TPM_OBJECT, bytes]): The value to encode

        Returns:
            Union[int, str, dict. list, None] depending on input value.

        Raises:
            TypeError: if val is either a TPM union or if type of val is unsupported.
        """
        if isinstance(val, TPM_BASE_OBJECT) and self._is_union(val):
            raise TypeError(f"tried to encode union {val.__class__.__name__}")
        if isinstance(val, TPMA_FRIENDLY_INTLIST):
            return self.encode_friendly_intlist(val)
        elif isinstance(val, TPM_FRIENDLY_INT):
            return self.encode_friendly_int(val)
        elif isinstance(val, int):
            return self.encode_int(val)
        elif isinstance(val, TPM2B_SIMPLE_OBJECT):
            return self.encode_simple_tpm2b(val)
        elif isinstance(
            val,
            (
                TPM2B_PUBLIC,
                TPM2B_SENSITIVE_CREATE,
                TPM2B_ECC_POINT,
                TPM2B_SENSITIVE,
                TPM2B_NV_PUBLIC,
                TPM2B_CREATION_DATA,
            ),
        ):
            return self.encode_complex_tpm2b(val)
        elif isinstance(val, TPML_OBJECT):
            return self.encode_tpml(val)
        elif isinstance(val, TPM_OBJECT):
            return self.encode_struct(val)
        elif isinstance(val, bytes):
            if len(val) == 0:
                return None
            return hexlify(val).decode("ascii")
        raise TypeError(f"unable to encode value of type {val.__class__.__name__}")

    def encode_int(self, val: int) -> int:
        """Encode an integer value

        Args:
            val (int): The value to encode

        Returns:
            The value as an int
        """
        return val

    def encode_friendly_int(self, val: TPM_FRIENDLY_INT) -> int:
        """Encode a TPM_FRIENDLY_INT value

        Args:
            val (TPM_FRIENDLY_INT): The value to encode

        Returns:
            The value as an int
        """
        return int(val)

    def encode_friendly_intlist(self, val: TPMA_FRIENDLY_INTLIST) -> int:
        """Encode a TPMA_FRIENDLY_INTLIST value

        Args:
            val (TPMA_FRIENDLY_INTLIST): The value to encode

        Returns:
            The value as an int
        """
        return int(val)

    def encode_simple_tpm2b(self, val: TPM2B_SIMPLE_OBJECT) -> Union[str, None]:
        """Encode a TPM2B_SIMPLE_OBJECT value

        Args:
            val (TPM2B_SIMPLE_OBJECT): The value to encode

        Returns:
            The value buffer as hex encoded string, or None if empty
        """
        if len(val) == 0:
            return None
        return str(val)

    def encode_complex_tpm2b(
        self,
        val: Union[
            TPM2B_PUBLIC,
            TPM2B_SENSITIVE_CREATE,
            TPM2B_ECC_POINT,
            TPM2B_SENSITIVE,
            TPM2B_NV_PUBLIC,
            TPM2B_CREATION_DATA,
        ],
    ) -> Dict[str, Any]:
        """Encode a complex TPM2B object value

        Args:
            val (Union[TPM2B_PUBLIC, TPM2B_SENSITIVE_CREATE, TPM2B_ECC_POINT, TPM2B_SENSITIVE, TPM2B_NV_PUBLIC, TPM2B_CREATION_DATA]): The value to encode

        Returns:
            The non-size field encoded as a dict
        """
        val = self._get_complex_field(val)
        return self.encode_struct(val)

    def encode_tpml(self, val: TPML_OBJECT) -> list:
        """Encode a TPML_OBJECT value

        Args:
            val (TPML_OBJECT): The value to encode

        Returns:
            A list with all the elements encoded
        """
        l = list()
        for v in val:
            ev = self.encode(v)
            l.append(ev)
        return l

    def encode_pcrselect(self, val: bytes) -> List[int]:
        """Encode a pcrSelect value

        Args:
            val (bytes): The value to encode

        Returns:
            A list containing all the set PCRs
        """
        pcrs = []
        val = reversed(bytes(val))
        si = int.from_bytes(val, "big")
        for i in range(0, si.bit_length()):
            b = si >> i
            if 1 & b:
                pcrs.append(i)

        return pcrs

    def encode_struct(self, val: TPM_OBJECT) -> Dict[str, Any]:
        """Encode a TPM_OBJECT value

        Args:
            val (TPM_OBJECT): The value to encode

        Returns:
            A dict containing all the encoded fields
        """
        d = dict()
        attrs = dir(val._cdata)
        for a in attrs:
            av = getattr(val, a)
            if self._is_union(av):
                av = self._get_by_selector(val, a)
                if av is None:
                    continue
            if a == "pcrSelect" and "sizeofSelect" in attrs:
                sv = self.encode_pcrselect(av[0 : val.sizeofSelect])
            elif a == "sizeofSelect":
                continue
            else:
                sv = self.encode(av)
            if sv is not None:
                d[a] = sv
        if len(d) == 0:
            return None
        return d

    def decode(
        self,
        dst: Union[
            TPMA_FRIENDLY_INTLIST,
            TPM_FRIENDLY_INT,
            int,
            TPM2B_SIMPLE_OBJECT,
            TPML_OBJECT,
            TPM_OBJECT,
        ],
        src: Union[int, str, dict, list],
    ) -> Union[
        TPMA_FRIENDLY_INTLIST,
        TPM_FRIENDLY_INT,
        int,
        TPM2B_SIMPLE_OBJECT,
        TPML_OBJECT,
        TPM_OBJECT,
    ]:
        """Decode a value into a TPM type

        Args:
            dst (Union[TPMA_FRIENDLY_INTLIST, TPM_FRIENDLY_INT, int, TPM2B_SIMPLE_OBJECT, TPML_OBJECT, TPM_OBJECT]): The (type) instance to decode into
            src (Union[int, str, dict, list]): The value to decode

        Returns:
            The decoded value as a Union[TPMA_FRIENDLY_INTLIST, TPM_FRIENDLY_INT, int, TPM2B_SIMPLE_OBJECT, TPML_OBJECT, TPM_OBJECT]

        Raises:
            TypeError: if dst is not a supported type or dst is a TPM union
        """
        if isinstance(dst, TPM_BASE_OBJECT) and self._is_union(dst):
            raise TypeError(f"tried to decode union {dst.__class__.__name__}")
        if isinstance(dst, TPMA_FRIENDLY_INTLIST):
            return self.decode_friendly_intlist(dst, src)
        elif isinstance(dst, TPM_FRIENDLY_INT):
            return self.decode_friendly_int(dst, src)
        elif isinstance(dst, int):
            return self.decode_int(dst, src)
        elif isinstance(dst, TPM2B_SIMPLE_OBJECT):
            return self.decode_simple_tpm2b(dst, src)
        elif isinstance(
            dst,
            (
                TPM2B_PUBLIC,
                TPM2B_SENSITIVE_CREATE,
                TPM2B_ECC_POINT,
                TPM2B_SENSITIVE,
                TPM2B_NV_PUBLIC,
                TPM2B_CREATION_DATA,
            ),
        ):
            return self.decode_complex_tpm2b(dst, src)
        elif isinstance(dst, TPML_OBJECT):
            return self.decode_tpml(dst, src)
        elif isinstance(dst, TPM_OBJECT):
            return self.decode_struct(dst, src)
        raise TypeError(f"unable to decode value of type {dst.__class__.__name__}")

    def decode_int(self, dst: int, src: int) -> int:
        """Decode an integer value

        Args:
            dst (int): the instance type to encode the integer as
            src (int): the int value to decode

        Returns:
            The int value as them same type as dst
        """
        return type(dst)(src)

    def decode_friendly_int(self, dst: TPM_FRIENDLY_INT, src: int) -> TPM_FRIENDLY_INT:
        """Decode a TPM_FRIENDLY_INT value

        Args:
            dst (TPM_FRIENDLY_INT): the instance type to encode the integer as
            src (int): the int value to decode

        Returns:
            The int value as them same type as dst
        """
        return type(dst)(src)

    def decode_friendly_intlist(
        self, dst: TPMA_FRIENDLY_INTLIST, src: int
    ) -> TPMA_FRIENDLY_INTLIST:
        """Decode a TPMA_FRIENDLY_INTLIST value

        Args:
            dst (TPMA_FRIENDLY_INTLIST): the instance type to encode the integer as
            src (int): the int value to decode

        Returns:
            The int value as them same type as dst
        """
        return type(dst)(src)

    def decode_simple_tpm2b(
        self, dst: TPM2B_SIMPLE_OBJECT, src: str
    ) -> TPM2B_SIMPLE_OBJECT:
        """Decode a TPM2B_SIMPLE_OBJECT value

        Args:
            dst (TPM2B_SIMPLE_OBJECT): the instance to decode into
            src (str): the value as a hex encoded string

        Returns:
            dst containing the decoded value
        """
        dst.buffer = unhexlify(src)
        return dst

    def decode_complex_tpm2b(
        self,
        dst: Union[
            TPM2B_PUBLIC,
            TPM2B_SENSITIVE_CREATE,
            TPM2B_ECC_POINT,
            TPM2B_SENSITIVE,
            TPM2B_NV_PUBLIC,
            TPM2B_CREATION_DATA,
        ],
        src: Dict[str, Any],
    ) -> Union[
        TPM2B_PUBLIC,
        TPM2B_SENSITIVE_CREATE,
        TPM2B_ECC_POINT,
        TPM2B_SENSITIVE,
        TPM2B_NV_PUBLIC,
        TPM2B_CREATION_DATA,
    ]:
        """Decode a complex TPM2V value

        Args:
            dst (Union[TPM2B_PUBLIC, TPM2B_SENSITIVE_CREATE, TPM2B_ECC_POINT, TPM2B_SENSITIVE, TPM2B_NV_PUBLIC, TPM2B_CREATION_DATA]): The instance to decode into
            src (dict): The value to decode

        Returns:
            dst containing the decoded value
        """
        f = self._get_complex_field(dst)
        f = self.decode(f, src)
        self._set_complex_field(dst, f)
        return dst

    def decode_tpml(self, dst: TPML_OBJECT, src: list) -> TPML_OBJECT:
        """Decode a TPML_OBJECT value

        Args:
            dst (TPML_OBJECT): The TPML_OBJECT to decode into
            src (list): the list of values to decode

        Returns:
            dst containing the decoded elements
        """
        ec = self._get_element_type(dst)
        i = 0
        for v in src:
            e = ec()
            e = self.decode(e, v)
            dst[i] = e
            i = i + 1
        return dst

    def decode_pcrselect(self, dst: bytes, src: List[int]) -> bytes:
        """Decode a pctSelect value

        Args:
            dst (bytes): the pcrSelect value, currently ignored
            src (List[int]): a list of the selected PCRs

        Returns:
            bytes with the selected PCR bits set
        """
        pcrs = 0
        for pcr in src:
            pcrs = pcrs | (1 << pcr)
        return bytes(reversed(pcrs.to_bytes(4, "big"))).rstrip(b"\x00")

    def decode_struct(self, dst: TPM_OBJECT, src: Dict[str, Any]) -> TPM_OBJECT:
        """Decode a TPM_OBJECT value

        Args:
            dst (TPM_OBJECT): The TPM_OBJECT instance to decode into
            src (dict): A dict of fields/values to decode

        Returns:
            dst containing the decoded values

        Raises:
            ValueError if strict is True and src containings unknown fields
        """
        if self._case_insensitive:
            fm = [(x.lower(), x) for x in dir(dst._cdata)]
            fields = dict(fm)
            km = [(x.lower(), x) for x in src.keys()]
            keys = dict(km)
        else:
            fm = [(x, x) for x in dir(dst._cdata)]
            fields = dict(fm)
            km = [(x, x) for x in src.keys()]
            keys = dict(km)
        mkeys = [x for x in keys if x in fields]
        if self._strict:
            missing = [x for x in keys if x not in fields]
            if len(missing) > 0:
                raise ValueError(f"unknown field(s) {', '.join(missing)} in source")
        for k in mkeys:
            rk = keys[k]
            rf = fields[k]
            df = getattr(dst, rf)
            if self._is_union(df):
                df = self._get_by_selector(dst, rf)
                rdst = df
            else:
                rdst = dst
            sf = src[rk]
            if rf == "pcrSelect" and "sizeofSelect" in fields.values():
                dv = self.decode_pcrselect(df, sf)
                setattr(rdst, "sizeofSelect", len(dv))
            elif rf == "sizeofSelect" and "pcrSelect" in fields.values():
                continue
            else:
                dv = self.decode(df, sf)
            setattr(rdst, rk, dv)
        return dst


class json_encdec(base_encdec):
    """Encode TPM types according to TCG TSS 2.0 JSON Data Types and Policy Language Specification

    Args:
        strict (bool): If a exception should be raised for unknown fields during decoding, defaults to True.
        case_insensitive (bool): If field names should be case insensitive during decoding, defaults to True.
    """

    def __init__(self, strict=True, case_insensitive=True):
        super().__init__(strict=strict, case_insensitive=case_insensitive)

    def encode_int(self, val: int) -> Union[List[int], int]:
        """Encode an int value

        Args:
            val (int): The value to encode

        Returns:
            An int or a list of two ints if size is over 54 bit
        """
        if val >= 0x100000000:
            return [(val >> 32) & 0xFFFFFFFF, val & 0xFFFFFFFF]
        return int(val)

    def encode_friendly_int(self, val: TPM_FRIENDLY_INT) -> Union[str, int, List[int]]:
        """Encode a TPM_FRIENDLY_INT value

        Args:
            val (TPM_FRIENDLY_INT): The value to encode

        Returns:
            A str if the val matches a constant, otherwise an encoded int
        """
        if isinstance(val, TPM_FRIENDLY_INT) and val in type(val):
            return str(val)
        return self.encode_int(val)

    def encode_friendly_intlist(self, val: TPMA_FRIENDLY_INTLIST) -> Dict[str, int]:
        """Encode a TPMA_FRIENDLY_INTLIST valut

        Args:
            val (TPMA_FRIENDLY_INT): The value to encode

        Returns
            A dict with the attribute names as the key and 1 as the value
        """
        attrs = dict()
        for i in range(0, 32):
            c = val & (1 << i)
            if c in type(val) and c != 0:
                k = str(c)
                attrs[k] = 1
        if isinstance(val, TPMA_NV) and val.nt:
            attrs["nt"] = self.encode_friendly_int(val.nt)
        if isinstance(val, TPMA_CC) and val.commandindex:
            attrs["commandindex"] = val.commandindex
        if isinstance(val, TPMA_CC) and val.chandles:
            attrs["chandles"] = val.chandles
        if isinstance(val, TPMA_LOCALITY) and val > 31:
            attrs["extended"] = (
                val & TPMA_LOCALITY.EXTENDED_MASK
            ) >> TPMA_LOCALITY.EXTENDED_SHIFT

        return attrs

    def decode_int(self, dst: int, src: Union[int, str, List[int]]) -> int:
        """Decode an int value

        Args:
            dst (int): The int instance, currently ignored
            src (Union[int, str, List[int]]): An int, a string containing the hex encoded value or a list of two ints

        Returns:
            A decoded int
        """
        if isinstance(src, str):
            dst = int(src, base=0)
        elif isinstance(src, list):
            dst = src[0] << 32 | src[1]
        else:
            dst = int(src)
        return dst

    def decode_friendly_int(
        self, dst: TPM_FRIENDLY_INT, src: Union[int, str]
    ) -> TPM_FRIENDLY_INT:
        """Decode a TPM_FRIENDLY_INT

        Args:
            dst (TPM_FRIENDLY_INT): The instance which type will be used
            src (Union[str, int]): Either a string containing the name of a constant or an int

        Returns:
            An instance of the same type as dst containing the decoded value
        """
        p = dst.__class__.__name__.lower() + "_"
        if isinstance(src, str):
            src = src.lower()
            if src.startswith(p):
                src = src[len(p)]
        try:
            return type(dst).parse(src)
        except (TypeError, ValueError):
            pass
        v = self.decode_int(dst, src)
        return type(dst)(v)

    def decode_friendly_intlist(
        self, dst: TPMA_FRIENDLY_INTLIST, src: Union[Dict[str, Any], List[str]]
    ) -> TPMA_FRIENDLY_INTLIST:
        """Decode a TPMA_FRIENDLY_INTLIST value

        Args:
            dst (TPMA_FRIENDLY_INTLIST): The instance which type will be used
            src (Union[Dict[str, Any], List[str]]): Either a dict with the attribute name as the key or a list of attributes

        Returns:
            An instance of the same type as dst containing the decoded value

        """
        attrs = type(dst)()
        p = dst.__class__.__name__.lower() + "_"
        if isinstance(src, dict):
            for k, v in src.items():
                k = k.lower()
                if k.startswith(p):
                    k = k[len(p) :]
                if isinstance(dst, TPMA_NV) and k == "nt":
                    nt = self.decode_friendly_int(TPM2_NT(), v)
                    attrs = attrs | nt << TPMA_NV.TPM2_NT_SHIFT
                    continue
                elif isinstance(dst, TPMA_CC) and k == "commandindex":
                    commandindex = self.decode_int(int(), v)
                    attrs = attrs | commandindex
                    continue
                elif isinstance(dst, TPMA_CC) and k == "chandles":
                    chandles = self.decode_int(int(), v)
                    attrs = attrs | chandles << TPMA_CC.CHANDLES_SHIFT
                    continue
                elif isinstance(dst, TPMA_LOCALITY) and k == "extended":
                    extended = self.decode_int(int(), v)
                    attrs = attrs | extended << TPMA_LOCALITY.EXTENDED_SHIFT
                    continue
                a = type(dst).parse(k)
                if (isinstance(v, str) and v.lower() in ("clear", "no")) or not v:
                    attrs = attrs & ~a
                else:
                    attrs = attrs | a
        elif isinstance(src, list):
            for v in src:
                v = v.lower()
                if v.startswith(p):
                    v = v[len(p) :]
                a = type(dst).parse(v)
                attrs = attrs | a
        else:
            attrs = self.decode_int(dst, src)

        return attrs

    def decode_simple_tpm2b(self, dst, src):
        """Decode a TPM2B_SIMPLE_OBJECT value

        Args:
            dst (TPM2B_SIMPLE_OBJECT): The object to store the decoded value in
            src (Union[str, List[int]]): Either a hex encoded string or a list of integers

        Returns:
            dst containing the decoded value
        """
        if isinstance(src, list):
            dst.buffer = bytes(src)
            return dst
        if src[0:2] == "0x":
            src = src[2:]
        dst.buffer = unhexlify(src)
        return dst



#!/usr/bin/python3 -u
# SPDX-License-Identifier: BSD-2
import unittest

from tpm2_pytss.internal.utils import TSS2Version


class InternalTest(unittest.TestCase):
    def test_lib_version_strings(self):

        versions = []

        # try the variants with a single major number
        versions.append(TSS2Version("1-rc0"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_00000000_00000000_00
        )

        versions.append(TSS2Version("1-rc0-dirty"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_00000000_00000000_01
        )

        versions.append(TSS2Version("1-rc0-2-g0fd1c5fbbaf2"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_00000000_00000002_00
        )

        versions.append(TSS2Version("1-rc0-2-g0fd1c5fbbaf2-dirty"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_00000000_00000002_01
        )

        versions.append(TSS2Version("1"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_FFFFFFFF_00000000_00
        )

        versions.append(TSS2Version("1-5-g0fd1c5fbbaf2"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_FFFFFFFF_00000005_00
        )

        versions.append(TSS2Version("1-5-g0fd1c5fbbaf2-dirty"))
        self.assertEqual(
            versions[-1], 0x00000001_00000000_00000000_FFFFFFFF_00000005_01
        )

        # try major minor variants
        versions.append(TSS2Version("2.0-rc0"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_00000000_00000000_00
        )

        versions.append(TSS2Version("2.0-rc0-dirty"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_00000000_00000000_01
        )

        versions.append(TSS2Version("2.0-rc0-2-g0fd1c5fbbaf2"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_00000000_00000002_00
        )

        versions.append(TSS2Version("2.0-rc0-2-g0fd1c5fbbaf2-dirty"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_00000000_00000002_01
        )

        versions.append(TSS2Version("2.0"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_FFFFFFFF_00000000_00
        )

        versions.append(TSS2Version("2.0-5-g0fd1c5fbbaf2"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_FFFFFFFF_00000005_00
        )

        versions.append(TSS2Version("2.0-5-g0fd1c5fbbaf2-dirty"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000000_FFFFFFFF_00000005_01
        )

        # try major minor patch variants
        versions.append(TSS2Version("2.0.1-rc0"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_00000000_00000000_00
        )

        versions.append(TSS2Version("2.0.1-rc0-dirty"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_00000000_00000000_01
        )

        versions.append(TSS2Version("2.0.1-rc0-2-g0fd1c5fbbaf2"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_00000000_00000002_00
        )

        versions.append(TSS2Version("2.0.1-rc0-2-g0fd1c5fbbaf2-dirty"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_00000000_00000002_01
        )

        versions.append(TSS2Version("2.0.1"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_FFFFFFFF_00000000_00
        )

        versions.append(TSS2Version("2.0.1-5-g0fd1c5fbbaf2"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_FFFFFFFF_00000005_00
        )

        versions.append(TSS2Version("2.0.1-5-g0fd1c5fbbaf2-dirty"))
        self.assertEqual(
            versions[-1], 0x00000002_00000000_00000001_FFFFFFFF_00000005_01
        )

        # We carefully built this list in sorted order so everything should be ascending
        # in value or precedence related to index. It value at index 5 is greater than
        # the value at index 2
        v0 = versions[0]
        for v1 in versions[1:]:
            self.assertTrue(v0 < v1, f"{str(v0)} not less than {str(v1)}")
            v0 = v1

        with self.assertRaises(ValueError, msg='Invalid version string, got: "1.N.0"'):
            TSS2Version("1.N.0")

        with self.assertRaises(ValueError, msg='Invalid version string, got: "N.1.0"'):
            TSS2Version("N.1.0")

        with self.assertRaises(ValueError, msg='Invalid version string, got: "3.1.N"'):
            TSS2Version("3.1.N")

        with self.assertRaises(
            ValueError, msg='Invalid version string, got: "3.1.0-N-g0fd1c5fbbaf2-dirty"'
        ):
            TSS2Version("3.1.0-N-g0fd1c5fbbaf2-dirty")

        with self.assertRaises(
            ValueError, msg='Invalid version string, got: "3.1.0-126-g0fd1c5fbbaf2-bad"'
        ):
            TSS2Version("3.1.0-126-g0fd1c5fbbaf2-bad")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# SPDX-License-Identifier: BSD-2

import os
import pkgconfig
import pathlib
import re
import sys
import textwrap


def remove_common_guards(s):

    # Remove includes and guards
    s = re.sub(r"#ifndef.*", r"", s)
    s = re.sub(r"#if .*", r"", s)
    s = re.sub(r"#define .*_H_*?\n", r"", s)
    s = re.sub(r"#endif.*", r"", s)
    s = re.sub(r"#error.*", r"", s)
    s = re.sub(r'#ifdef __cplusplus\nextern "C" {', r"", s, flags=re.MULTILINE)
    s = re.sub(r"#ifdef __cplusplus\n}", r"", s, flags=re.MULTILINE)
    s = re.sub(r"#include.*", "", s)

    # Remove certain macros
    s = re.sub(r"#define TSS2_API_VERSION.*", r"", s)
    s = re.sub(r"#define TSS2_ABI_VERSION.*", r"", s)
    s = re.sub(r"#define TSS2_RC_LAYER\(level\).*", r"", s)
    s = re.sub(r"(#define.*)TSS2_RC_LAYER\(0xff\)", r"\g<1>0xff0000", s)

    # Remove comments
    s = re.sub(r"/\*.*?\*/", "", s, flags=re.MULTILINE)

    # Restructure #defines with ...
    s = re.sub(r"(#define [A-Za-z0-9_]+) +\(\(.*?\) \(.*?\)\)", r"\g<1>...", s)
    s = re.sub(r"(#define [A-Za-z0-9_]+) +\(\(\(.*?\) .*\)", r"\g<1>...", s)
    s = re.sub(r"(#define [A-Za-z0-9_]+) +\(\(.*?\).*?\) ", r"\g<1>...", s)
    s = re.sub(
        r"(#define [A-Za-z0-9_]+) .*\n.*?.*\)\)", r"\g<1>...", s, flags=re.MULTILINE
    )
    s = re.sub(r"(#define +[A-Za-z0-9_]+) +\((-?\d+)\)", r"\g<1> \g<2>", s)
    s = re.sub(r"(#define [A-Za-z0-9_]+) .*", r"\g<1>...", s)

    # Restructure structs and untions with ...
    s = re.sub(r"\[.+?\]", r"[...]", s)

    return s


def remove_poll_stuff(s, poll_handle_type):

    r = r"#if defined\(__linux__\).*(\n.*)+#endif\n#endif"

    s = re.sub(r, rf"typedef struct pollfd {poll_handle_type};", s)
    return s


def remove_INTERNALBUILD(s):

    r = r"#if\s+defined\(INTERNALBUILD\)(?:(?!endif).)*#endif"
    s = re.sub(r, "", s, flags=re.MULTILINE | re.DOTALL)
    s = re.sub(r"DEPRECATED", r"", s)

    return re.sub(r"__attribute__\(\(deprecated\)\)", r"", s)


def prepare_common(dirpath):

    s = pathlib.Path(dirpath, "tss2_common.h").read_text(encoding="utf-8")

    return remove_common_guards(s)


def prepare_types(dirpath):

    s = pathlib.Path(dirpath, "tss2_tpm2_types.h").read_text(encoding="utf-8")

    # Remove false define (workaround)
    s = re.sub(
        r"#define TPM2_MAX_TAGGED_POLICIES.*\n.*TPMS_TAGGED_POLICY\)\)",
        r"",
        s,
        flags=re.MULTILINE,
    )

    s = remove_INTERNALBUILD(s)

    s = re.sub(
        r"typedef struct TPMS_ALGORITHM_DESCRIPTION TPMS_ALGORITHM_DESCRIPTION ;",
        r"",
        s,
    )

    s = re.sub(
        r"struct TPMS_ALGORITHM_DESCRIPTION {\n.*\n.*\n} ;", r"", s, flags=re.MULTILINE
    )

    return remove_common_guards(s)


def prepare_tcti(dirpath):

    s = pathlib.Path(dirpath, "tss2_tcti.h").read_text(encoding="utf-8")

    s = re.sub(r"#ifndef TSS2_API_VERSION.*\n.*\n#endif", r"", s, flags=re.MULTILINE)

    s = remove_poll_stuff(s, "TSS2_TCTI_POLL_HANDLE")

    s = re.sub(r"#define TSS2_TCTI_.*\n.*", r"", s, flags=re.MULTILINE)
    s = re.sub(r"^\s*#define Tss2_Tcti_(?:.*\\\r?\n)*.*$", r"", s, flags=re.MULTILINE)

    s += """
    struct pollfd {
        int   fd;         /* file descriptor */
        short events;     /* requested events */
        short revents;    /* returned events */
    };
    """

    # Add the callbacks for a TCTI
    s += """
    extern "Python" TSS2_RC _tcti_transmit_wrapper (
        TSS2_TCTI_CONTEXT *tctiContext,
        size_t size,
        uint8_t const *command);

    extern "Python" TSS2_RC _tcti_receive_wrapper (
        TSS2_TCTI_CONTEXT *tctiContext,
        size_t *size,
        uint8_t *response,
        int32_t timeout);

    extern "Python" void _tcti_finalize_wrapper (
        TSS2_TCTI_CONTEXT *tctiContext);

    extern "Python" TSS2_RC _tcti_cancel_wrapper (
        TSS2_TCTI_CONTEXT *tctiContext);

    extern "Python" TSS2_RC _tcti_get_pollfds_wrapper (
    TSS2_TCTI_CONTEXT *tctiContext,
    TSS2_TCTI_POLL_HANDLE *handles,
    size_t *num_handles);

    extern "Python" TSS2_RC _tcti_set_locality_wrapper (
        TSS2_TCTI_CONTEXT *tctiContext,
        uint8_t locality);

    extern "Python" TSS2_RC _tcti_make_sticky_wrapper (
        TSS2_TCTI_CONTEXT *tctiContext,
        TPM2_HANDLE *handle,
        uint8_t sticky);
    """

    # Add this struct here so CFFI knows about it and its size
    s += """
    typedef struct PYTCTI_CONTEXT PYTCTI_CONTEXT;
    struct PYTCTI_CONTEXT {
        TSS2_TCTI_CONTEXT_COMMON_V2 common;
        void *thiz;
    };
    """

    return remove_common_guards(s)


def prepare_tcti_ldr(dirpath):

    s = pathlib.Path(dirpath, "tss2_tctildr.h").read_text(encoding="utf-8")
    return remove_common_guards(s)


def prepare_tcti_spi_helper(dirpath):

    s = pathlib.Path(dirpath, "tss2_tcti_spi_helper.h").read_text(encoding="utf-8")

    # Add the callbacks for a TCTI
    s += """
    extern "Python" TSS2_RC _tcti_spi_helper_sleep_ms (
        void *userdata,
        int milliseconds);

    extern "Python" TSS2_RC _tcti_spi_helper_start_timeout (
        void *userdata,
        int milliseconds);

    extern "Python" TSS2_RC _tcti_spi_helper_timeout_expired (
        void *userdata, bool *is_timeout_expired);

    extern "Python" TSS2_RC _tcti_spi_helper_spi_acquire (
        void *userdata);

    extern "Python" TSS2_RC _tcti_spi_helper_spi_release (
    void *userdata);

    extern "Python" TSS2_RC _tcti_spi_helper_spi_transfer (
        void *userdata,
        const void *data_out,
        void *data_in,
        size_t cnt);

    extern "Python" void _tcti_spi_helper_finalize (
        void *userdata);
    """

    return remove_common_guards(s)


def prepare_sapi():
    return "typedef struct TSS2_SYS_CONTEXT TSS2_SYS_CONTEXT;"


def prepare_esapi(dirpath):

    s = pathlib.Path(dirpath, "tss2_esys.h").read_text(encoding="utf-8")
    return remove_common_guards(s)


def prepare_fapi(dirpath):

    s = pathlib.Path(dirpath, "tss2_fapi.h").read_text(encoding="utf-8")

    s = remove_poll_stuff(s, "FAPI_POLL_HANDLE")

    s += """
    extern "Python" TSS2_RC _fapi_auth_callback(
        char     const *objectPath,
        char     const *description,
        char    const **auth,
        void           *userData);
    extern "Python" TSS2_RC _fapi_branch_callback(
        char     const *objectPath,
        char     const *description,
        char    const **branchNames,
        size_t          numBranches,
        size_t         *selectedBranch,
        void           *userData);
    extern "Python" TSS2_RC _fapi_sign_callback(
        char     const *objectPath,
        char     const *description,
        char     const *publicKey,
        char     const *publicKeyHint,
        uint32_t        hashAlg,
        uint8_t  const *dataToSign,
        size_t          dataToSignSize,
        uint8_t const **signature,
        size_t         *signatureSize,
        void           *userData);
    extern "Python" TSS2_RC _fapi_policy_action_callback(
        char     const *objectPath,
        char     const *action,
        void           *userData);
     """

    return remove_common_guards(s)


def prepare_rcdecode(dirpath):

    s = pathlib.Path(dirpath, "tss2_rc.h").read_text(encoding="utf-8")

    return remove_common_guards(s)


def prepare_mu(dirpath):

    s = pathlib.Path(dirpath, "tss2_mu.h").read_text(encoding="utf-8")

    s = remove_INTERNALBUILD(s)

    s = remove_common_guards(s)

    # At least tpm2-tss 3.0.3 have duplicated BYTE (un)marshal functions which break cffi
    # So removing them is needed until 3.1.x has reached most distributions
    n = re.findall(
        r"TSS2_RC\s+Tss2_MU_BYTE_Marshal\(.+?\);", s, re.DOTALL | re.MULTILINE
    )
    if len(n) > 1:
        s = re.sub(
            r"TSS2_RC\s+Tss2_MU_BYTE_Marshal\(.+?\);",
            r"",
            s,
            1,
            re.DOTALL | re.MULTILINE,
        )

    n = re.findall(
        r"TSS2_RC\s+Tss2_MU_BYTE_Unmarshal\(.+?\);", s, re.DOTALL | re.MULTILINE
    )
    if len(n) > 1:
        s = re.sub(
            r"TSS2_RC\s+Tss2_MU_BYTE_Unmarshal\(.+?\);",
            r"",
            s,
            1,
            re.DOTALL | re.MULTILINE,
        )

    s = re.sub(
        r"TSS2_RC\s+Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Marshal\(.+?\)\s+;",
        r"",
        s,
        1,
        re.DOTALL | re.MULTILINE,
    )

    s = re.sub(
        r"TSS2_RC\s+Tss2_MU_TPMS_ALGORITHM_DESCRIPTION_Unmarshal\(.+?\)\s+;",
        "",
        s,
        1,
        re.DOTALL | re.MULTILINE,
    )

    return s


def prepare_policy(dirpath):
    s = pathlib.Path(dirpath, "tss2_policy.h").read_text(encoding="utf-8")
    s = remove_common_guards(s)
    # cparser complains if a typedef of an enum is before the definition of the enum
    s = re.sub(
        r"typedef enum TSS2_POLICY_PCR_SELECTOR TSS2_POLICY_PCR_SELECTOR;", r"", s, 1
    )
    s = re.sub(
        r"(enum TSS2_POLICY_PCR_SELECTOR.*?\};)",
        r"\1" + "\ntypedef enum TSS2_POLICY_PCR_SELECTOR TSS2_POLICY_PCR_SELECTOR;",
        s,
        1,
        re.DOTALL | re.MULTILINE,
    )
    s += """
    extern "Python" TSS2_RC _policy_cb_calc_pcr(
        TSS2_POLICY_PCR_SELECTION *selection,
        TPML_PCR_SELECTION *out_selection,
        TPML_DIGEST *out_digest,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_calc_name(
        const char *path,
        TPM2B_NAME *name,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_calc_public(
        const char *path,
        TPMT_PUBLIC *public,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_calc_nvpublic(
        const char *path,
        TPMI_RH_NV_INDEX nv_index,
        TPMS_NV_PUBLIC *nv_public,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_auth(
        TPM2B_NAME *name,
        ESYS_TR *object_handle,
        ESYS_TR *auth_handle,
        ESYS_TR *authSession,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_polsel(
        TSS2_OBJECT *auth_object,
        const char **branch_names,
        size_t branch_count,
        size_t *branch_idx,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_sign(
        char *key_pem,
        char *public_key_hint,
        TPMI_ALG_HASH key_pem_hash_alg,
        uint8_t *buffer,
        size_t buffer_size,
        const uint8_t **signature,
        size_t *signature_size,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_polauth(
        TPMT_PUBLIC *key_public,
        TPMI_ALG_HASH hash_alg,
        TPM2B_DIGEST *digest,
        TPM2B_NONCE *policyRef,
        TPMT_SIGNATURE *signature,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_polauthnv(
        TPMS_NV_PUBLIC *nv_public,
        TPMI_ALG_HASH hash_alg,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_poldup(
        TPM2B_NAME *name,
        void *userdata);
    extern "Python" TSS2_RC _policy_cb_exec_polaction(
        const char *action,
        void *userdata);
    """
    return s


def prepare(
    indir, outfile, build_fapi=True, build_policy=True, build_tcti_spi_helper=True
):
    indir = os.path.join(indir, "tss2")

    common = prepare_common(indir)

    types = prepare_types(indir)

    tcti = prepare_tcti(indir)

    tcti_ldr = prepare_tcti_ldr(indir)

    if build_tcti_spi_helper:
        tcti_spi_helper = prepare_tcti_spi_helper(indir)

    sapi = prepare_sapi()

    esapi = prepare_esapi(indir)

    if build_fapi:
        fapi = prepare_fapi(indir)

    rcdecode = prepare_rcdecode(indir)

    mu = prepare_mu(indir)

    if build_policy:
        policy = prepare_policy(indir)

    # Write result
    with open(outfile, "w") as f:
        f.write(
            textwrap.dedent(
                """
            /*
             * SPDX-License-Identifier: BSD-2
             * This file was automatically generated. Do not modify !
             */
            """
            )
        )

        f.write(common)
        f.write(types)
        f.write(tcti)
        f.write(tcti_ldr)
        if build_tcti_spi_helper:
            f.write(tcti_spi_helper)
        f.write(sapi)
        f.write(esapi)
        if build_fapi:
            f.write(fapi)
        f.write(rcdecode)
        f.write(mu)
        if build_policy:
            f.write(policy)


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: {0} <tss2-header-dir> <output-file>".format(sys.argv[0]))
        exit(1)

    build_fapi = pkgconfig.installed("tss2-fapi", ">=3.0.0")
    build_tcti_spi_helper = pkgconfig.exists("tss2-tcti-spi-helper")
    prepare(
        sys.argv[1],
        sys.argv[2],
        build_fapi=build_fapi,
        build_tcti_spi_helper=build_tcti_spi_helper,
    )



# SPDX-License-Identifier: BSD-2

import shutil
import logging
import os
import random
import socket
import subprocess
import sys
import tempfile
import time
import unittest

from tpm2_pytss import *


class BaseTpmSimulator(object):
    def __init__(self):
        self.tpm = None
        self._port = None

    @staticmethod
    def ready(port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(("localhost", port)) == 0

    def start(self):
        logger = logging.getLogger("DEBUG")
        logger.debug('Setting up simulator: "{}"'.format(self.exe))

        tpm = None
        for _ in range(0, 20):

            random_port = random.randrange(2321, 65534)

            sim = self._start(port=random_port)
            for _ in range(0, 10):
                rc = sim.poll()
                if rc is not None:
                    logger.debug(f"Simulator {self.exe} exited with {rc}")
                    break
                if (
                    sim.poll() is None
                    and self.ready(random_port)
                    and self.ready(random_port + 1)
                ):
                    tpm = sim
                    break
                time.sleep(0.1)

            if tpm:
                self.tpm = sim
                self._port = random_port
                logger.debug(f"started {self.exe} on port {random_port}\n")
                break
            else:
                sim.kill()

        if not tpm:
            raise SystemError("Could not start simulator")

    def close(self):
        if self.tpm.poll() is not None:
            return
        self.tpm.terminate()
        try:
            self.tpm.wait(timeout=1)
        except subprocess.TimeoutExpired:
            self.tpm.kill()
        self.tpm.wait(timeout=10)

    def __str__(self):
        return self.exe


class SwtpmSimulator(BaseTpmSimulator):
    exe = "swtpm"
    libname = "libtss2-tcti-swtpm.so"

    def __init__(self):
        super().__init__()
        self.working_dir = tempfile.TemporaryDirectory()

    def _start(self, port):

        cmd = [
            "swtpm",
            "socket",
            "--tpm2",
            "--server",
            "port={}".format(port),
            "--ctrl",
            "type=tcp,port={}".format(port + 1),
            "--flags",
            "not-need-init",
            "--tpmstate",
            "dir={}".format(self.working_dir.name),
        ]

        tpm = subprocess.Popen(cmd)
        return tpm

    @property
    def tcti_name_conf(self):
        if self._port is None:
            return None

        return f"swtpm:port={self._port}"

    def get_tcti(self):
        if self._port is None:
            return None

        return TCTILdr("swtpm", f"port={self._port}")


class IBMSimulator(BaseTpmSimulator):
    exe = "tpm_server"
    libname = "libtss2-tcti-mssim.so"

    def __init__(self):
        super().__init__()
        self.working_dir = tempfile.TemporaryDirectory()

    def _start(self, port):

        cwd = os.getcwd()
        os.chdir(self.working_dir.name)
        try:
            cmd = ["tpm_server", "-rm", "-port", "{}".format(port)]
            tpm = subprocess.Popen(cmd)
            return tpm

        finally:
            os.chdir(cwd)

    @property
    def tcti_name_conf(self):
        if self._port is None:
            return None

        return f"mssim:port={self._port}"

    def get_tcti(self):
        if self._port is None:
            return None

        return TCTILdr("mssim", f"port={self._port}")


class TpmSimulator(object):

    SIMULATORS = [
        SwtpmSimulator,
        IBMSimulator,
    ]

    @staticmethod
    def getSimulator():

        for sim in TpmSimulator.SIMULATORS:
            exe = shutil.which(sim.exe)
            if not exe:
                print(f'Could not find executable: "{sim.exe}"', file=sys.stderr)
                continue
            if not TCTILdr.is_available(sim.libname):
                continue

            return sim()

        raise RuntimeError(
            "Expected to find a TPM 2.0 Simulator, tried {}, got None".format(
                TpmSimulator.SIMULATORS
            )
        )


class TSS2_BaseTest(unittest.TestCase):
    def setUp(self):
        self.tpm = TpmSimulator.getSimulator()
        self.tpm.start()

    def tearDown(self):
        self.tpm.close()


class TSS2_EsapiTest(TSS2_BaseTest):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tcti = None
        self.ectx = None

    def skipIfAlgNotSupported(self, alg: TPM2_ALG):
        self.skipTest(f'Algorithm "{alg}" not supported by simulator')

    def skipIfCommandNotSupported(self, cc: TPM2_CC):
        self.skipTest(f'Command "{cc}" not supported by simulator')

    def setUp(self):
        super().setUp()
        try:
            # use str initializer here to test string inputs to ESAPI constructor
            with ESAPI(self.tpm.tcti_name_conf) as ectx:
                ectx.startup(TPM2_SU.CLEAR)

        except Exception as e:
            self.tpm.close()
            raise e
        self.tcti = self.tpm.get_tcti()
        self.ectx = ESAPI(self.tcti)

        # record the supported algorithms
        self._supported_algs = []
        more = True
        while more:
            more, data = self.ectx.get_capability(
                TPM2_CAP.ALGS, 0, lib.TPM2_MAX_CAP_ALGS
            )
            self._supported_algs += [x.alg for x in data.data.algorithms]

        self._supported_commands = []
        more = True
        while more:
            more, data = self.ectx.get_capability(TPM2_CAP.ALGS, 0, lib.TPM2_MAX_CAP_CC)
            self._supported_commands += list(data.data.command)

    def tearDown(self):
        self.ectx.close()
        self.tcti.close()
        super().tearDown()
